import os

from qtpy.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QGroupBox,
    QCheckBox,
    QLabel,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QGridLayout,
    QTextEdit,
)
from qtpy.QtCore import Qt

from .useful_widgets import LineEditReadOnly, PushButtonMinimumWidth, set_tooltip

import logging

logger = logging.getLogger(__name__)


class DialogExportToTiffAndTxt(QDialog):
    def __init__(self, parent=None, *, dir_path=""):

        super().__init__(parent)

        self.__dir_path = ""
        self.__save_tiff = True
        self.__save_txt = False
        self.__interpolate_on = False
        self.__quant_norm_on = False
        self.__dset_list = []
        self.__dset_sel = 0
        self.__scaler_list = []
        self.__scaler_sel = 0

        self.setWindowTitle("Export XRF Maps to TIFF and/or TXT Files")
        self.setMinimumHeight(600)
        self.setMinimumWidth(600)
        self.resize(600, 600)

        self.te_saved_files = QTextEdit()
        set_tooltip(self.te_saved_files, "The list of <b>data file groups</b> about to be created.")
        self.te_saved_files.setReadOnly(True)

        self.combo_select_dataset = QComboBox()
        self.combo_select_dataset.currentIndexChanged.connect(self.combo_select_dataset_current_index_changed)
        self._fill_dataset_combo()
        set_tooltip(
            self.combo_select_dataset,
            "Select <b>dataset</b>. Initially, the selection matches the dataset activated "
            "in <b>XRF Maps</b> tab, but the selection may be changed if different dataset "
            "needs to be saved.",
        )

        self.combo_normalization = QComboBox()
        self.combo_normalization.currentIndexChanged.connect(self.combo_normalization_current_index_changed)
        self._fill_scaler_combo()
        set_tooltip(
            self.combo_normalization,
            "Select <b>scaler</b> used for data normalization. Initially, the selection matches "
            "the scaler activated in <b>XRF Maps</b> tab, but the selection may be changed "
            "if needed",
        )

        self.cb_interpolate = QCheckBox("Interpolate to uniform grid")
        self.cb_interpolate.setChecked(Qt.Checked if self.__interpolate_on else Qt.Unchecked)
        self.cb_interpolate.stateChanged.connect(self.cb_interpolate_state_changed)
        set_tooltip(
            self.cb_interpolate,
            "Interpolate pixel coordinates to <b>uniform grid</b>. The initial choice is "
            "copied from <b>XRF Maps</b> tab.",
        )

        self.cb_quantitative = QCheckBox("Quantitative normalization")
        self.cb_quantitative.setChecked(Qt.Checked if self.__quant_norm_on else Qt.Unchecked)
        self.cb_quantitative.stateChanged.connect(self.cb_quantitative_state_changed)
        set_tooltip(
            self.cb_quantitative,
            "Apply <b>quantitative normalization</b> before saving the maps. "
            "The initial choice is copied from <b>XRF Maps</b> tab.",
        )

        self.group_settings = QGroupBox("Settings (selections from XRF Maps tab)")
        grid = QGridLayout()
        grid.addWidget(self.combo_select_dataset, 0, 0)
        grid.addWidget(self.combo_normalization, 0, 1)
        grid.addWidget(self.cb_interpolate, 1, 0)
        grid.addWidget(self.cb_quantitative, 1, 1)
        self.group_settings.setLayout(grid)

        self.le_dir_path = LineEditReadOnly()
        set_tooltip(
            self.le_dir_path,
            "<b>Root directory</b> for saving TIFF and TXT files. The files will be saved "
            "in subdirectories inside the root directory.",
        )
        self.pb_dir_path = PushButtonMinimumWidth("..")
        set_tooltip(self.pb_dir_path, "Change to <b>root directory</b> for TIFF and TXT files.")
        self.pb_dir_path.clicked.connect(self.pb_dir_path_clicked)
        self.pb_dir_path.setDefault(False)
        self.pb_dir_path.setAutoDefault(False)

        self.cb_save_tiff = QCheckBox("Save TIFF")
        set_tooltip(self.cb_save_tiff, "Save XRF Maps as <b>TIFF</b> files.")
        self.cb_save_tiff.setChecked(Qt.Checked if self.__save_tiff else Qt.Unchecked)
        self.cb_save_tiff.stateChanged.connect(self.cb_save_tiff_state_changed)
        self.cb_save_txt = QCheckBox("Save TXT")
        self.cb_save_txt.setChecked(Qt.Checked if self.__save_txt else Qt.Unchecked)
        set_tooltip(self.cb_save_txt, "Save XRF Maps as <b>TXT</b> files.")
        self.cb_save_txt.stateChanged.connect(self.cb_save_txt_state_changed)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.group_settings)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("The following groups of files will be created:"))
        hbox.addStretch(1)
        vbox.addLayout(hbox)
        vbox.addWidget(self.te_saved_files)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Directory: "))
        hbox.addWidget(self.pb_dir_path)
        hbox.addWidget(self.le_dir_path)
        vbox.addLayout(hbox)

        self.button_box = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        self.pb_save = self.button_box.button(QDialogButtonBox.Save)
        self.pb_save.setDefault(False)
        self.pb_save.setAutoDefault(False)
        self.pb_cancel = self.button_box.button(QDialogButtonBox.Cancel)
        self.pb_cancel.setDefault(True)
        self.pb_cancel.setAutoDefault(True)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self.cb_save_tiff)
        hbox.addWidget(self.cb_save_txt)
        hbox.addStretch(1)
        hbox.addWidget(self.button_box)
        vbox.addLayout(hbox)

        self.setLayout(vbox)

        # Set and display file path
        dir_path = os.path.expanduser(dir_path)
        self.dir_path = dir_path

        self._update_pb_save()
        self._update_saved_file_groups()

    @property
    def dir_path(self):
        return self.__dir_path

    @dir_path.setter
    def dir_path(self, dir_path):
        self.__dir_path = dir_path
        self.le_dir_path.setText(dir_path)

    @property
    def save_tiff(self):
        return self.__save_tiff

    @save_tiff.setter
    def save_tiff(self, save_tiff):
        self.__save_tiff = save_tiff
        self.cb_save_tiff.setChecked(Qt.Checked if save_tiff else Qt.Unchecked)
        self._update_pb_save()

    @property
    def save_txt(self):
        return self.__save_txt

    @save_txt.setter
    def save_txt(self, save_txt):
        self.__save_txt = save_txt
        self.cb_save_txt.setChecked(Qt.Checked if save_txt else Qt.Unchecked)
        self._update_pb_save()

    @property
    def interpolate_on(self):
        return self.__interpolate_on

    @interpolate_on.setter
    def interpolate_on(self, interpolate_on):
        self.__interpolate_on = interpolate_on
        self.cb_interpolate.setChecked(Qt.Checked if interpolate_on else Qt.Unchecked)

    @property
    def quant_norm_on(self):
        return self.__quant_norm_on

    @quant_norm_on.setter
    def quant_norm_on(self, quant_norm_on):
        self.__quant_norm_on = quant_norm_on
        self.cb_quantitative.setChecked(Qt.Checked if quant_norm_on else Qt.Unchecked)

    @property
    def dset_list(self):
        return self.__dset_list

    @dset_list.setter
    def dset_list(self, dset_list):
        self.__dset_list = dset_list
        self._fill_dataset_combo()

    @property
    def dset_sel(self):
        return self.__dset_sel

    @dset_sel.setter
    def dset_sel(self, dset_sel):
        self.__dset_sel = dset_sel
        self.combo_select_dataset.setCurrentIndex(dset_sel - 1)

    @property
    def scaler_list(self):
        return self.__scaler_list

    @scaler_list.setter
    def scaler_list(self, scaler_list):
        self.__scaler_list = scaler_list
        self._fill_scaler_combo()

    @property
    def scaler_sel(self):
        return self.__scaler_sel

    @scaler_sel.setter
    def scaler_sel(self, scaler_sel):
        self.__scaler_sel = scaler_sel
        self.combo_normalization.setCurrentIndex(scaler_sel)

    def pb_dir_path_clicked(self):
        # Note: QFileDialog.ShowDirsOnly is not set on purpose, so that the dialog
        #   could be used to inspect directory contents. Files can not be selected anyway.
        dir_path = QFileDialog.getExistingDirectory(
            self,
            "Select Root Directory for TIFF and TXT Files",
            self.dir_path,
            options=QFileDialog.DontResolveSymlinks,
        )
        if dir_path:
            self.dir_path = dir_path

    def _update_pb_save(self):
        state = self.__save_tiff or self.__save_txt
        self.pb_save.setEnabled(state)

    def cb_save_tiff_state_changed(self, state):
        self.__save_tiff = state == Qt.Checked
        self._update_pb_save()
        self._update_saved_file_groups()

    def cb_save_txt_state_changed(self, state):
        self.__save_txt = state == Qt.Checked
        self._update_pb_save()
        self._update_saved_file_groups()

    def cb_interpolate_state_changed(self, state):
        self.__interpolate_on = state == Qt.Checked

    def cb_quantitative_state_changed(self, state):
        self.__quant_norm_on = state == Qt.Checked
        self._update_saved_file_groups()

    def combo_select_dataset_current_index_changed(self, index):
        self.__dset_sel = index + 1
        self._update_saved_file_groups()

    def combo_normalization_current_index_changed(self, index):
        self.__scaler_sel = index
        self._update_saved_file_groups()

    def get_selected_dset_name(self):
        index = self.__dset_sel - 1
        n = len(self.__dset_list)
        if index < 0 or index >= n:
            return None
        else:
            return self.__dset_list[index]

    def get_selected_scaler_name(self):
        index = self.__scaler_sel - 1
        n = len(self.__scaler_list)
        if index < 0 or index >= n:
            return None
        else:
            return self.__scaler_list[index]

    def _fill_dataset_combo(self):
        self.combo_select_dataset.clear()
        self.combo_select_dataset.addItems(self.__dset_list)
        self.combo_select_dataset.setCurrentIndex(self.__dset_sel - 1)

    def _fill_scaler_combo(self):
        scalers = ["Normalize by ..."] + self.__scaler_list
        self.combo_normalization.clear()
        self.combo_normalization.addItems(scalers)
        self.combo_normalization.setCurrentIndex(self.__scaler_sel)

    def _update_saved_file_groups(self):
        dset_name = self.get_selected_dset_name()
        scaler_name = self.get_selected_scaler_name()
        is_fit, is_roi = False, False
        if dset_name is not None:
            is_fit = dset_name.endswith("fit")
            is_roi = dset_name.endswith("roi")

        text_common = ""
        if is_fit:
            text_common += "  - Fitted XRF maps\n"
        elif is_roi:
            text_common += "  - ROI maps\n"
        if (is_fit or is_roi) and (scaler_name is not None):
            text_common += f"  - Normalized maps (scaler '{scaler_name}')\n"
        if is_fit and self.__quant_norm_on:
            text_common += "  - Quantitative maps (if calibration data is loaded)\n"
        text_common += "  - Scalers\n"
        text_common += "  - Positional coordinates\n"

        text = ""
        if self.__save_tiff:
            text += "TIFF FORMAT:\n" + text_common
        if self.__save_txt:
            if text:
                text += "\n"
            text += "TXT_FORMAT:\n" + text_common
        if not text:
            text = "No files will be saved"
        self.te_saved_files.setText(text)



import copy
import os
import numpy as np
import numpy.testing as npt
import pytest
from pyxrf.api_dev import save_data_to_hdf5, read_data_from_hdf5


def _prepare_raw_dataset(N=5, M=10, K=4096):

    det1 = np.ones(shape=[N, M, K]) * 100
    det2 = np.ones(shape=[N, M, K]) * 150
    det3 = np.ones(shape=[N, M, K]) * 200
    det_sum = det1 + det2 + det3

    scaler_names = ["i0", "scaler1", "scaler2"]
    scaler_data = np.ones(shape=[N, M, len(scaler_names)])
    for n in range(len(scaler_names)):
        scaler_data[:, :, n] *= n

    pos_names = ["x_pos", "y_pos"]
    pos_data = np.zeros(shape=[2, N, M])
    pos_data[0, :, :] = np.broadcast_to(np.linspace(1, 1 + (M - 1) * 0.1, M), shape=[N, M])
    pos_data[1, :, :] = np.broadcast_to(
        np.reshape(np.linspace(5, 5 + (N - 1) * 0.2, N), newshape=[N, 1]), shape=[N, M]
    )

    data = {
        "det_sum": det_sum,
        "det1": det1,
        "det2": det2,
        "det3": det3,
        "scaler_names": scaler_names,
        "scaler_data": scaler_data,
        "pos_names": pos_names,
        "pos_data": pos_data,
    }

    metadata = {"scan_id": 1000, "scan_uid": "SOME-RANDOM-UID-STRING", "name": "Some more metadata"}

    return data, metadata


def test_save_data_to_hdf5_1(tmp_path):
    """
    Basic test for ``save_data_to_hdf5``
    """
    fln = "test.h5"
    fpath = os.path.join(tmp_path, fln)
    data, metadata = _prepare_raw_dataset(N=5, M=10, K=4096)

    save_data_to_hdf5(fpath, data, metadata=metadata)

    # Check if the file exists
    assert os.path.isfile(fpath)

    # Check that the exception occurs in case of attempt to overwrite the file
    with pytest.raises(IOError, match="File .* already exists"):
        save_data_to_hdf5(fpath, data, metadata=metadata)

    # File should be overwritten
    save_data_to_hdf5(fpath, data, metadata=metadata, file_overwrite_existing=True)

    # Different version of the file should be created
    save_data_to_hdf5(fpath, data, metadata=metadata, fname_add_version=True)
    save_data_to_hdf5(fpath, data, metadata=metadata, fname_add_version=True)

    # Check if the file exists
    assert os.path.isfile(fpath)
    assert os.path.isfile(os.path.join(tmp_path, "test_v1.h5"))
    assert os.path.isfile(os.path.join(tmp_path, "test_v2.h5"))


# fmt: off
@pytest.mark.parametrize("sets_to_select, sets_to_save", [
    ("all", "default"),
    ("all", "all"),
    ("channels", "all"),
    ("all", "sum"),
    ("channels", "sum"),
    ("sum", "sum"),
    ("sum", "default"),
    ("sum", "all"),
])
# fmt: on
def test_save_data_to_hdf5_2(tmp_path, sets_to_select, sets_to_save):
    """
    Save data to hdf5 and then read it.
    """
    fln = "test.h5"
    fpath = os.path.join(tmp_path, fln)
    data, metadata = _prepare_raw_dataset(N=5, M=10, K=4096)

    data_to_save = copy.deepcopy(data)
    if sets_to_select == "sum":
        # Leave only
        del data_to_save["det1"]
        del data_to_save["det2"]
        del data_to_save["det3"]
    elif sets_to_select == "channels":
        # Leave only sum
        del data_to_save["det_sum"]

    kwargs = {}
    if sets_to_save == "all":
        kwargs = {"create_each_det": True}
    elif sets_to_save == "sum":
        kwargs = {"create_each_det": False}

    save_data_to_hdf5(fpath, data_to_save, metadata=metadata, **kwargs)
    data_loaded, metadata_loaded = read_data_from_hdf5(fpath)

    metadata_selected = {_: metadata_loaded[_] for _ in metadata}
    assert metadata_selected == metadata

    assert data_loaded["pos_names"] == data["pos_names"]
    npt.assert_array_almost_equal(data_loaded["pos_data"], data["pos_data"])

    assert isinstance(data_loaded["det_sum"], np.ndarray)
    npt.assert_array_almost_equal(data_loaded["det_sum"], data["det_sum"])
    if (sets_to_select == "sum") or (sets_to_save == "sum"):
        # Only 'sum' of all channels will be saved
        keys = set(data.keys())
        keys.remove("det1")
        keys.remove("det2")
        keys.remove("det3")
        assert "det1" not in data_loaded
        assert "det2" not in data_loaded
        assert "det3" not in data_loaded
        assert set(data_loaded.keys()) == keys
    else:
        assert set(data_loaded.keys()) == set(data.keys())
        npt.assert_array_almost_equal(data_loaded["det1"], data["det1"])
        npt.assert_array_almost_equal(data_loaded["det2"], data["det2"])
        npt.assert_array_almost_equal(data_loaded["det3"], data["det3"])

    assert data_loaded["pos_names"] == data["pos_names"]
    npt.assert_array_almost_equal(data_loaded["pos_data"], data["pos_data"])
    assert data_loaded["scaler_names"] == data["scaler_names"]
    npt.assert_array_almost_equal(data_loaded["scaler_data"], data["scaler_data"])


def test_save_data_to_hdf5_3(tmp_path):
    """
    Verify that default metadata is automatically created.
    """
    fln = "test.h5"
    fpath = os.path.join(tmp_path, fln)
    data, metadata = _prepare_raw_dataset(N=5, M=10, K=4096)

    save_data_to_hdf5(fpath, data)
    data_loaded, metadata_loaded = read_data_from_hdf5(fpath)

    metadata_keys = [
        "file_type",
        "file_format",
        "file_format_version",
        "file_created_time",
        "file_software",
        "file_software_version",
    ]
    for k in metadata_keys:
        assert k in metadata_loaded


def test_save_data_to_hdf5_4(tmp_path):
    """
    Verify that metadata ``file_software`` and ``file_software_version`` can be set by the user.
    """
    fln = "test.h5"
    fpath = os.path.join(tmp_path, fln)
    data, metadata = _prepare_raw_dataset(N=5, M=10, K=4096)

    application = "Testing Application"
    version = "v4.5.6"

    metadata = {"file_software": application, "file_software_version": version}

    save_data_to_hdf5(fpath, data, metadata=metadata)
    data_loaded, metadata_loaded = read_data_from_hdf5(fpath)

    assert metadata_loaded["file_software"] == application
    assert metadata_loaded["file_software_version"] == version



import os

from qtpy.QtWidgets import (
    QPushButton,
    QHBoxLayout,
    QVBoxLayout,
    QGroupBox,
    QCheckBox,
    QComboBox,
    QListWidget,
    QListWidgetItem,
    QDialog,
    QFileDialog,
    QMessageBox,
)
from qtpy.QtCore import Qt, Signal, Slot, QThreadPool, QRunnable

from .useful_widgets import (
    LineEditReadOnly,
    adjust_qlistwidget_height,
    global_gui_parameters,
    PushButtonMinimumWidth,
    set_tooltip,
    clear_gui_state,
)
from .form_base_widget import FormBaseWidget
from .dlg_load_mask import DialogLoadMask
from .dlg_select_scan import DialogSelectScan
from .dlg_view_metadata import DialogViewMetadata
import logging

logger = logging.getLogger(__name__)


class LoadDataWidget(FormBaseWidget):

    update_main_window_title = Signal()
    update_global_state = Signal()
    computations_complete = Signal(object)

    update_preview_map_range = Signal(str)
    signal_new_run_loaded = Signal(bool)  # True/False - success/failed
    signal_loading_new_run = Signal()  # Emitted before new run is loaded

    signal_data_channel_changed = Signal(bool)

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.ref_main_window = self.gui_vars["ref_main_window"]

        self.update_global_state.connect(self.ref_main_window.update_widget_state)

        self.initialize()

    def initialize(self):

        v_spacing = global_gui_parameters["vertical_spacing_in_tabs"]

        vbox = QVBoxLayout()

        self._setup_wd_group()
        vbox.addWidget(self.group_wd)
        vbox.addSpacing(v_spacing)

        self._setup_load_group()
        vbox.addWidget(self.group_file)
        vbox.addSpacing(v_spacing)

        self._setup_sel_channel_group()
        vbox.addWidget(self.group_sel_channel)
        vbox.addSpacing(v_spacing)

        self._setup_spec_settings_group()
        vbox.addWidget(self.group_spec_settings)
        vbox.addSpacing(v_spacing)

        self._setup_preview_group()
        vbox.addWidget(self.group_preview)

        vbox.addStretch(1)

        self.setLayout(vbox)

        self._set_tooltips()

    def _setup_wd_group(self):
        self.group_wd = QGroupBox("Working Directory")

        self.pb_set_wd = PushButtonMinimumWidth("..")
        self.pb_set_wd.clicked.connect(self.pb_set_wd_clicked)

        self.le_wd = LineEditReadOnly()

        # Initial working directory. Set to the HOME directory for now
        current_dir = os.path.expanduser(self.gpc.get_current_working_directory())
        self.le_wd.setText(current_dir)

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_set_wd)
        hbox.addWidget(self.le_wd)

        self.group_wd.setLayout(hbox)

    def _setup_load_group(self):

        self.group_file = QGroupBox("Load Data")

        self.pb_file = QPushButton("Read File ...")
        self.pb_file.clicked.connect(self.pb_file_clicked)

        self.pb_dbase = QPushButton("Load Run ...")
        self.pb_dbase.setEnabled(self.gui_vars["gui_state"]["databroker_available"])
        self.pb_dbase.clicked.connect(self.pb_dbase_clicked)

        self.cb_file_all_channels = QCheckBox("All channels")
        self.cb_file_all_channels.setChecked(self.gpc.get_load_each_channel())
        self.cb_file_all_channels.toggled.connect(self.cb_file_all_channels_toggled)

        self.le_file_default = "No data is loaded"
        self.le_file = LineEditReadOnly(self.le_file_default)

        self.pb_view_metadata = QPushButton("View Metadata ...")
        self.pb_view_metadata.setEnabled(False)
        self.pb_view_metadata.clicked.connect(self.pb_view_metadata_clicked)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_file)
        hbox.addWidget(self.pb_dbase)
        hbox.addWidget(self.cb_file_all_channels)
        vbox.addLayout(hbox)

        vbox.addWidget(self.le_file)

        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self.pb_view_metadata)
        vbox.addLayout(hbox)

        self.group_file.setLayout(vbox)

    def _setup_sel_channel_group(self):

        self.group_sel_channel = QGroupBox("Select Channel For Processing")

        self.cbox_channel = QComboBox()
        self.cbox_channel.currentIndexChanged.connect(self.cbox_channel_index_changed)
        self._set_cbox_channel_items(items=[])

        hbox = QHBoxLayout()
        hbox.addWidget(self.cbox_channel)

        self.group_sel_channel.setLayout(hbox)

    def _setup_spec_settings_group(self):

        self.group_spec_settings = QGroupBox("Total Spectrum Settings")

        self.pb_apply_mask = QPushButton("Apply Mask ...")
        self.pb_apply_mask.clicked.connect(self.pb_apply_mask_clicked)

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_apply_mask)
        hbox.addStretch(1)

        self.group_spec_settings.setLayout(hbox)

    def _setup_preview_group(self):

        self.group_preview = QGroupBox("Preview")

        self.list_preview = QListWidget()
        self.list_preview.itemChanged.connect(self.list_preview_item_changed)
        self.list_preview.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.list_preview.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self._set_list_preview_items(items=[])

        hbox = QHBoxLayout()
        hbox.addWidget(self.list_preview)

        self.group_preview.setLayout(hbox)

    def _set_tooltips(self):
        set_tooltip(
            self.pb_set_wd,
            "Select <b>Working Directory</b>. The Working Directory is "
            "used as <b>default</b> for loading and saving data and "
            "configuration files.",
        )
        set_tooltip(self.le_wd, "Currently selected <b>Working Directory</b>")
        set_tooltip(self.pb_file, "Load data from a <b>file on disk</b>.")
        set_tooltip(self.pb_dbase, "Load data from a <b>database</b> (Databroker).")
        set_tooltip(
            self.cb_file_all_channels,
            "Load <b>all</b> available data channels (checked) or only the <b>sum</b> of the channels",
        )
        set_tooltip(self.le_file, "The <b>name</b> of the loaded file or <b>ID</b> of the loaded run.")
        set_tooltip(self.pb_view_metadata, "View scan <b>metadata</b> (if available)")
        set_tooltip(self.cbox_channel, "Select channel for processing. Typically the <b>sum</b> channel is used.")
        set_tooltip(
            self.pb_apply_mask,
            "Load the mask from file and/or select spatial ROI. The mask and ROI "
            "are used in run <b>Preview</b> and fitting of the <b>total spectrum</b>.",
        )
        set_tooltip(
            self.list_preview,
            "Data for the selected channels is displayed in <b>Preview</b> tab. "
            "The displayed <b>total spectrum</b> is computed for the selected "
            "ROI and using the loaded mask. If no mask or ROI are enabled, then "
            "total spectrum is computed over all pixels of the image.",
        )

    def update_widget_state(self, condition=None):
        if condition == "tooltips":
            self._set_tooltips()

        state = self.gui_vars["gui_state"]["state_file_loaded"]
        self.group_sel_channel.setEnabled(state)
        self.group_spec_settings.setEnabled(state)
        self.group_preview.setEnabled(state)

    def _set_cbox_channel_items(self, *, items=None):
        """
        Set items of the combo box for item selection. If the list of items is not specified,
        then it is loaded from the respective data structure.

        Parameters
        ----------
        items: list(str)
            The list of items. The list may be cleared by calling
            `self._set_cbox_channel_items(items=[])`

        """
        self.cbox_channel.currentIndexChanged.disconnect(self.cbox_channel_index_changed)

        self.cbox_channel.clear()
        if items is None:
            items = list(self.gpc.get_file_channel_list())
        self.cbox_channel.addItems(items)

        self.cbox_channel.currentIndexChanged.connect(self.cbox_channel_index_changed)

        if len(items):
            # Select the first item (if there is at least one item)
            self.cbox_channel.setCurrentIndex(0)

    def _set_list_preview_items(self, *, items=None):
        """
        Set items of the list for selecting channels in preview tab. If the list of items is
        not specified, then it is loaded from the respective data structure.

        Parameters
        ----------
        items: list(str)
            The list of items. The list may be cleared by calling
            `self._set_cbox_channel_items(items=[])`

        """
        self.list_preview.itemChanged.disconnect(self.list_preview_item_changed)

        self.list_preview.clear()
        if items is None:
            items = list(self.gpc.get_file_channel_list())
        for s in items:
            wi = QListWidgetItem(s, self.list_preview)
            wi.setFlags(wi.flags() | Qt.ItemIsUserCheckable)
            wi.setFlags(wi.flags() & ~Qt.ItemIsSelectable)
            wi.setCheckState(Qt.Unchecked)

        # Adjust height so that it fits all the elements
        adjust_qlistwidget_height(self.list_preview, other_widgets=[self.group_preview, self])

        # This will cause the preview data to be plotted (the plot is expected to be hidden,
        #   since no channels were selected). Here we select the first channel in the list.
        for n, item in enumerate(items):
            state = Qt.Checked if self.gpc.is_dset_item_selected_for_preview(item) else Qt.Unchecked
            self.list_preview.item(n).setCheckState(state)

        self.list_preview.itemChanged.connect(self.list_preview_item_changed)

    def pb_set_wd_clicked(self):
        dir_current = self.le_wd.text()
        dir = QFileDialog.getExistingDirectory(
            self,
            "Select Working Directory",
            dir_current,
            QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
        )
        if dir:
            self.gpc.set_current_working_directory(dir)
            self.le_wd.setText(dir)

    def pb_file_clicked(self):
        dir_current = self.gpc.get_current_working_directory()
        file_paths = QFileDialog.getOpenFileName(self, "Open Data File", dir_current, "HDF5 (*.h5);; All (*)")
        file_path = file_paths[0]
        if file_path:
            self.signal_loading_new_run.emit()

            def cb(file_path):
                result_dict = {}
                try:
                    msg = self.gpc.open_data_file(file_path)
                    status = True
                except Exception as ex:
                    msg = str(ex)
                    status = False
                result_dict.update({"status": status, "msg": msg, "file_path": file_path})
                return result_dict

            self._compute_in_background(cb, self.slot_file_clicked, file_path=file_path)

    @Slot(object)
    def slot_file_clicked(self, result):
        self._recover_after_compute(self.slot_file_clicked)

        status = result["status"]
        msg = result["msg"]  # Message is empty if file loading failed
        file_path = result["file_path"]
        if status:
            file_text = f"'{self.gpc.get_loaded_file_name()}'"
            if self.gpc.is_scan_metadata_available():
                file_text += f": ID#{self.gpc.get_metadata_scan_id()}"
            self.le_file.setText(file_text)

            self.gui_vars["gui_state"]["state_file_loaded"] = True
            # Invalidate fit. Fit must be rerun for new data.
            self.gui_vars["gui_state"]["state_model_fit_exists"] = False
            # Check if any datasets were loaded.
            self.gui_vars["gui_state"]["state_xrf_map_exists"] = self.gpc.is_xrf_maps_available()

            # Disable the button for changing working directory. This is consistent
            #   with the behavior of the old PyXRF, but will be changed in the future.
            self.pb_set_wd.setEnabled(False)

            # Enable/disable 'View Metadata' button
            self.pb_view_metadata.setEnabled(self.gpc.is_scan_metadata_available())
            self.le_wd.setText(self.gpc.get_current_working_directory())

            self.update_main_window_title.emit()
            self.update_global_state.emit()

            self._set_cbox_channel_items()
            self._set_list_preview_items()

            # Here we want to reset the range in the Total Count Map preview
            self.update_preview_map_range.emit("reset")

            self.signal_new_run_loaded.emit(True)  # Data is loaded successfully

            if msg:
                # Display warning message if it was generated
                msgbox = QMessageBox(QMessageBox.Warning, "Warning", msg, QMessageBox.Ok, parent=self)
                msgbox.exec()

        else:
            self.le_file.setText(self.le_file_default)

            # Disable 'View Metadata' button
            self.pb_view_metadata.setEnabled(False)
            self.le_wd.setText(self.gpc.get_current_working_directory())

            # Clear flags: the state now is "No data is loaded".
            clear_gui_state(self.gui_vars)
            self.update_global_state.emit()

            self.update_main_window_title.emit()
            self.update_global_state.emit()

            self._set_cbox_channel_items(items=[])
            self._set_list_preview_items(items=[])

            # Here we want to clear the range in the Total Count Map preview
            self.update_preview_map_range.emit("clear")

            self.signal_new_run_loaded.emit(False)  # Failed to load data

            msg_str = (
                f"Incorrect format of input file '{file_path}': "
                f"PyXRF accepts only custom HDF (.h5) files."
                f"\n\nError message: {msg}"
            )
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg_str, QMessageBox.Ok, parent=self)
            msgbox.exec()

    def pb_dbase_clicked(self):

        dlg = DialogSelectScan()
        if dlg.exec() == QDialog.Accepted:
            mode, id_uid = dlg.get_id_uid()

            self.signal_loading_new_run.emit()

            def cb(id_uid):
                result_dict = {}
                try:
                    msg, file_name = self.gpc.load_run_from_db(id_uid)
                    status = True
                except Exception as ex:
                    msg, file_name = str(ex), ""
                    status = False
                result_dict.update({"status": status, "msg": msg, "id_uid": id_uid, "file_name": file_name})
                return result_dict

            self._compute_in_background(cb, self.slot_dbase_clicked, id_uid=id_uid)

    @Slot(object)
    def slot_dbase_clicked(self, result):
        self._recover_after_compute(self.slot_dbase_clicked)

        status = result["status"]
        msg = result["msg"]  # Message is empty if file loading failed
        id_uid = result["id_uid"]
        # file_name = result["file_name"]
        if status:
            file_text = f"'{self.gpc.get_loaded_file_name()}'"
            if self.gpc.is_scan_metadata_available():
                file_text += f": ID#{self.gpc.get_metadata_scan_id()}"
            self.le_file.setText(file_text)

            self.gui_vars["gui_state"]["state_file_loaded"] = True
            # Invalidate fit. Fit must be rerun for new data.
            self.gui_vars["gui_state"]["state_model_fit_exists"] = False
            # Check if any datasets were loaded.
            self.gui_vars["gui_state"]["state_xrf_map_exists"] = self.gpc.is_xrf_maps_available()

            # Disable the button for changing working directory. This is consistent
            #   with the behavior of the old PyXRF, but will be changed in the future.
            self.pb_set_wd.setEnabled(False)

            # Enable/disable 'View Metadata' button
            self.pb_view_metadata.setEnabled(self.gpc.is_scan_metadata_available())
            self.le_wd.setText(self.gpc.get_current_working_directory())

            self.update_main_window_title.emit()
            self.update_global_state.emit()

            self._set_cbox_channel_items()
            self._set_list_preview_items()

            # Here we want to reset the range in the Total Count Map preview
            self.update_preview_map_range.emit("reset")

            self.signal_new_run_loaded.emit(True)  # Data is loaded successfully

            if msg:
                # Display warning message if it was generated
                msgbox = QMessageBox(QMessageBox.Warning, "Warning", msg, QMessageBox.Ok, parent=self)
                msgbox.exec()

        else:
            self.le_file.setText(self.le_file_default)

            # Disable 'View Metadata' button
            self.pb_view_metadata.setEnabled(False)
            self.le_wd.setText(self.gpc.get_current_working_directory())

            # Clear flags: the state now is "No data is loaded".
            clear_gui_state(self.gui_vars)
            self.update_global_state.emit()

            self.update_main_window_title.emit()
            self.update_global_state.emit()

            self._set_cbox_channel_items(items=[])
            self._set_list_preview_items(items=[])

            # Here we want to clear the range in the Total Count Map preview
            self.update_preview_map_range.emit("clear")

            self.signal_new_run_loaded.emit(False)  # Failed to load data

            msg_str = f"Failed to load scan '{id_uid}'.\n\nError message: {msg}"
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg_str, QMessageBox.Ok, parent=self)
            msgbox.exec()

    def pb_apply_mask_clicked(self):
        map_size = self.gpc.get_dataset_map_size()
        map_size = map_size if (map_size is not None) else (0, 0)

        dlg = DialogLoadMask()
        dlg.set_image_size(n_rows=map_size[0], n_columns=map_size[1])
        roi = self.gpc.get_preview_spatial_roi()
        dlg.set_roi(
            row_start=roi["row_start"],
            column_start=roi["col_start"],
            row_end=roi["row_end"],
            column_end=roi["col_end"],
        )
        dlg.set_roi_active(self.gpc.is_roi_selection_active())

        dlg.set_default_directory(self.gpc.get_current_working_directory())
        dlg.set_mask_file_path(self.gpc.get_mask_selection_file_path())
        dlg.set_mask_file_active(self.gpc.is_mask_selection_active())

        if dlg.exec() == QDialog.Accepted:
            roi_keys = ("row_start", "col_start", "row_end", "col_end")
            roi_list = dlg.get_roi()
            roi_selected = {k: roi_list[n] for n, k in enumerate(roi_keys)}
            self.gpc.set_preview_spatial_roi(roi_selected)

            self.gpc.set_roi_selection_active(dlg.get_roi_active())

            self.gpc.set_mask_selection_file_path(dlg.get_mask_file_path())
            self.gpc.set_mask_selection_active(dlg.get_mask_file_active())

            def cb():
                try:
                    # TODO: proper error processing is needed here (exception RuntimeError)
                    self.gpc.apply_mask_to_datasets()
                    success = True
                    msg = ""
                except Exception as ex:
                    success = False
                    msg = str(ex)
                return {"success": success, "msg": msg}

            self._compute_in_background(cb, self.slot_apply_mask_clicked)

    @Slot(object)
    def slot_apply_mask_clicked(self, result):
        if not result["success"]:
            msg = f"Error occurred while applying the ROI selection:\nException: {result['msg']}"
            logger.error(f"{msg}")
            mb_error = QMessageBox(QMessageBox.Critical, "Error", f"{msg}", QMessageBox.Ok, parent=self)
            mb_error.exec()

        # Here we want to expand the range in the Total Count Map preview if needed
        self.update_preview_map_range.emit("update")
        self._recover_after_compute(self.slot_apply_mask_clicked)

    def pb_view_metadata_clicked(self):

        dlg = DialogViewMetadata()
        metadata_string = self.gpc.get_formatted_metadata()
        dlg.setText(metadata_string)
        dlg.exec()

    def cb_file_all_channels_toggled(self, state):
        self.gpc.set_load_each_channel(state)

    def cbox_channel_index_changed(self, index):
        def cb(index):
            try:
                self.gpc.set_data_channel(index)
                success, msg = True, ""
            except Exception as ex:
                success = False
                msg = str(ex)
            return {"success": success, "msg": msg}

        self._compute_in_background(cb, self.slot_channel_index_changed, index=index)

    @Slot(object)
    def slot_channel_index_changed(self, result):
        self._recover_after_compute(self.slot_channel_index_changed)

        if result["success"]:
            self.signal_data_channel_changed.emit(True)
        else:
            self.signal_data_channel_changed.emit(False)
            msg = result["msg"]
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()

    def list_preview_item_changed(self, list_item):
        # Find the index of the list item that was checked/unchecked
        ind = -1
        for n in range(self.list_preview.count()):
            if self.list_preview.item(n) == list_item:
                ind = n

        # Get the state of the checkbox (checked -> 2, unchecked -> 0)
        state = list_item.checkState()

        # The name of the dataset
        dset_name = self.gpc.get_file_channel_list()[ind]

        def cb():
            try:
                self.gpc.select_preview_dataset(dset_name=dset_name, is_visible=bool(state))
                success, msg = True, ""
            except Exception as ex:
                success = False
                msg = str(ex)
            return {"success": success, "msg": msg}

        self._compute_in_background(cb, self.slot_preview_items_changed)

    @Slot(object)
    def slot_preview_items_changed(self, result):
        if not result["success"]:
            # The error shouldn't actually happen here. This is to prevent potential crashes.
            msg = f"Error occurred: {result['msg']}.\nData may need to be reloaded to continue processing."
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()

        # Here we want to expand the range in the Total Count Map preview if needed
        self.update_preview_map_range.emit("expand")
        self._recover_after_compute(self.slot_preview_items_changed)

    def _compute_in_background(self, func, slot, *args, **kwargs):
        """
        Run function `func` in a background thread. Send the signal
        `self.computations_complete` once computation is finished.

        Parameters
        ----------
        func: function
            Reference to a function that is supposed to be executed at the background.
            The function return value is passed as a signal parameter once computation is
            complete.
        slot: qtpy.QtCore.Slot or None
            Reference to a slot. If not None, then the signal `self.computation_complete`
            is connected to this slot.
        args, kwargs
            arguments of the function `func`.
        """
        signal_complete = self.computations_complete

        def func_to_run(func, *args, **kwargs):
            class LoadFile(QRunnable):
                def run(self):
                    result_dict = func(*args, **kwargs)
                    signal_complete.emit(result_dict)

            return LoadFile()

        if slot is not None:
            self.computations_complete.connect(slot)
        self.gui_vars["gui_state"]["running_computations"] = True
        self.update_global_state.emit()
        QThreadPool.globalInstance().start(func_to_run(func, *args, **kwargs))

    def _recover_after_compute(self, slot):
        """
        The function should be called after the signal `self.computations_complete` is
        received. The slot should be the same as the one used when calling
        `self.compute_in_background`.
        """
        if slot is not None:
            self.computations_complete.disconnect(slot)
        self.gui_vars["gui_state"]["running_computations"] = False
        self.update_global_state.emit()



# basic command line functions to use. Some of those functions can be implemented to
# scikit-beam later.
import os
import time
import json
import glob
import multiprocessing
import numpy as np
import h5py
import copy
from collections.abc import Iterable
import re

from ..core.quant_analysis import ParamQuantitativeAnalysis

from skbeam.core.fitting.xrf_model import linear_spectrum_fitting, define_range
from .fileio import output_data, read_hdf_APS, read_MAPS, sep_v
from .fit_spectrum import single_pixel_fitting_controller, save_fitdata_to_hdf

from ..core.map_processing import dask_client_create

import logging

logger = logging.getLogger(__name__)


def fit_pixel_data_and_save(
    working_directory,
    file_name,
    *,
    param_file_name,
    fit_channel_sum=True,
    fit_channel_each=False,
    param_channel_list=None,
    incident_energy=None,
    ignore_datafile_metadata=False,
    fln_quant_calib_data=None,
    quant_distance_to_sample=0,
    method="nnls",
    pixel_bin=0,
    raise_bg=0,
    comp_elastic_combine=False,
    linear_bg=False,
    use_snip=True,
    bin_energy=0,
    save_txt=False,
    save_tiff=True,
    scaler_name=None,
    use_average=False,
    interpolate_to_uniform_grid=False,
    data_from="NSLS-II",
    dask_client=None,
):
    """
    Do fitting for signle data set, and save data accordingly. Fitting can be performed on
    either summed data or each channel data, or both.

    Parameters
    ----------
    working_directory : str, required
        path folder
    file_names : str, required
        selected h5 file
    param_file_name : str, required
        param file name for summed data fitting
    fit_channel_sum : bool, optional
        fit summed data or not
    fit_channel_each : bool, optional
        fit each channel data or not
    param_channel_list : list, optional
        list of param file names for each channel
    incident_energy : float, optional
        use this energy as incident energy instead of the one in param file, i.e., XANES
        This value overrides the incident energy from metadata and from JSON parameter file
    ignore_datafile_metadata : bool
        tells whether to ignore metadata from the data file (if data file contains metadata).
        At the moment, only incident energy (monochromator energy) is used by the processing routine.
        If True, then the incident energy from the parameter JSON file is used; if False, then
        the energy from metadata is used. If the function parameter ``incident_energy`` is
        not None, then its value overrides the incident energy from metadata or JSON file.
    fln_quant_calib_data : str or list(str)
        file name or a list of file names that contain quantitative calibration data
    quant_distance_to_sample : float
        distance-to-sample used in quantitative calibration. If 0, then correction for
        distance is not applied (assumed that the standard and the sample were placed
        at the same distance from the detector).
    method : str, optional
        fitting method, default as nnls
    pixel_bin : int, optional
        bin pixel as 2by2, or 3by3
    raise_bg : int, optional
        add a constant value to each spectrum, better for fitting
    comp_elastic_combine : bool, optional
        combine elastic and compton as one component for fitting
    linear_bg : bool, optional
        use linear background instead of snip
    use_snip : bool, optional
        use snip method to remove background
    bin_energy : int, optional
        bin spectrum with given value
    save_txt : bool, optional
        save data to txt or not
    save_tiff : bool, optional
        save data to tiff or not
    scaler_name : str, optional
        name of the field representing the scaler (for example 'i0'), which must be present
        in the data files. If given, normalization will be performed before saving data
        as .txt and .tiff
    use_average : bool, optional
        if true, norm is performed as data/scaler*mean(scaler), otherwise just data/scaler
    interpolate_to_uniform_grid : bool
        interpolate the result to uniform grid before saving to tiff and txt files
        The grid dimensions match the dimensions of positional data for X and Y axes.
        The range of axes is chosen to fit the values of X and Y.
    data_from : str, optional
        where do data come from? Data format includes data from NSLS-II, or 2IDE-APS
    dask_client: dask.distributed.Client
        Dask client object. If None, then Dask client is created automatically.
        If a batch of files is processed, then creating Dask client and
        passing the reference to it to the processing functions will save
        execution time: `client = Client(processes=True, silence_logs=logging.ERROR)`
    """
    fpath = os.path.join(working_directory, file_name)

    # Load quantitative calibration files (if necessary)
    quant_norm = False  # Indicates if at least one calibration file is loaded
    param_quant_analysis = ParamQuantitativeAnalysis()
    if fln_quant_calib_data:
        if isinstance(fln_quant_calib_data, str):
            fln_quant_calib_data = [fln_quant_calib_data]
        for fln in fln_quant_calib_data:
            if os.path.isabs(fln):
                f = fln
            else:
                f = os.path.join(working_directory, fln)
            try:
                param_quant_analysis.load_entry(f)
                quant_norm = True
                logger.info(f"Quantitative calibration is loaded successfully from file '{f}'")
            except Exception as ex:
                logger.error(f"Error occurred while loading quantitative calibration from file '{f}': {ex}")

    t0 = time.time()
    prefix_fname = file_name.split(".")[0]
    if fit_channel_sum is True:
        if data_from == "NSLS-II":
            img_dict, data_sets, mdata = read_hdf_APS(working_directory, file_name, load_each_channel=False)
        elif data_from == "2IDE-APS":
            img_dict, data_sets, mdata = read_MAPS(working_directory, file_name, channel_num=1)
        else:
            print("Unknown data sets.")

        try:
            data_all_sum = data_sets[prefix_fname + "_sum"].raw_data
        except KeyError:
            data_all_sum = data_sets[prefix_fname].raw_data

        # load param file
        if not os.path.isabs(param_file_name):
            param_path = os.path.join(working_directory, param_file_name)
        else:
            param_path = param_file_name
        with open(param_path, "r") as json_data:
            param_sum = json.load(json_data)

        # update incident energy, required for XANES
        if incident_energy is not None:
            param_sum["coherent_sct_energy"]["value"] = incident_energy
            print("Using incident beam energy passed as the function parameter.")
        elif (
            mdata.is_metadata_available()
            and "instrument_mono_incident_energy" in mdata
            and not ignore_datafile_metadata
        ):
            param_sum["coherent_sct_energy"]["value"] = mdata["instrument_mono_incident_energy"]
            print(f"Using incident beam energy from the data file '{file_name}'.")
        else:
            print(f"Using incident beam energy from the parameter file '{param_path}'.")

        # The value of incident energy that is used for processing
        incident_energy_used = param_sum["coherent_sct_energy"]["value"]
        print(f"Incident beam energy: {incident_energy_used}.")

        result_map_sum, calculation_info = single_pixel_fitting_controller(
            data_all_sum,
            param_sum,
            incident_energy=incident_energy,
            method=method,
            pixel_bin=pixel_bin,
            raise_bg=raise_bg,
            comp_elastic_combine=comp_elastic_combine,
            linear_bg=linear_bg,
            use_snip=use_snip,
            bin_energy=bin_energy,
            dask_client=dask_client,
        )

        # output to .h5 file
        inner_path = "xrfmap/detsum"
        # fit_name = prefix_fname+'_fit'
        save_fitdata_to_hdf(fpath, result_map_sum, datapath=inner_path)

        def get_scaler_set(img_dict):
            sc_set_names = [_ for _ in img_dict if _.endswith("_scaler")]
            if sc_set_names:
                return img_dict[sc_set_names[0]]
            else:
                return {}

        def get_positions_set(img_dict):
            if "positions" in img_dict:
                return img_dict["positions"]
            else:
                return {}

        scaler_dict = get_scaler_set(img_dict)
        scaler_name_list = list(scaler_dict.keys())
        positions_dict = get_positions_set(img_dict)
        # Generate dataset
        dataset = copy.deepcopy(scaler_dict)
        dataset.update(result_map_sum)

        # Set parameters for quantitative normalization
        param_quant_analysis.experiment_incident_energy = incident_energy_used
        param_quant_analysis.experiment_distance_to_sample = quant_distance_to_sample
        param_quant_analysis.experiment_detector_channel = "sum"

        if save_txt is True:
            output_folder = "output_txt_" + prefix_fname
            output_path = os.path.join(working_directory, output_folder)
            output_data(
                output_dir=output_path,
                interpolate_to_uniform_grid=interpolate_to_uniform_grid,
                dataset_name="dataset_fit",  # Sum of all detectors: should end with '_fit'
                quant_norm=quant_norm,
                param_quant_analysis=param_quant_analysis,
                distance_to_sample=quant_distance_to_sample,
                dataset_dict=dataset,
                positions_dict=positions_dict,
                file_format="txt",
                scaler_name=scaler_name,
                scaler_name_list=scaler_name_list,
                use_average=use_average,
            )

        if save_tiff is True:
            output_folder = "output_tiff_" + prefix_fname
            output_path = os.path.join(working_directory, output_folder)
            output_data(
                output_dir=output_path,
                interpolate_to_uniform_grid=interpolate_to_uniform_grid,
                dataset_name="dataset_fit",  # Sum of all detectors: should end with '_fit'
                quant_norm=quant_norm,
                param_quant_analysis=param_quant_analysis,
                dataset_dict=dataset,
                positions_dict=positions_dict,
                file_format="tiff",
                scaler_name=scaler_name,
                scaler_name_list=scaler_name_list,
                use_average=use_average,
            )

    if fit_channel_each:
        img_dict, data_sets, mdata = read_hdf_APS(working_directory, file_name, load_each_channel=True)

        # Find the detector channels and the names of the channels
        det_channels = [_ for _ in data_sets.keys() if re.search(r"_det\d+$", _)]
        det_channel_names = [re.search(r"det\d+$", _)[0] for _ in det_channels]
        if param_channel_list is None:
            param_channel_list = [param_file_name] * 3
        else:
            if not isinstance(param_channel_list, list) and not isinstance(param_channel_list, tuple):
                raise RuntimeError("Parameter 'param_channel_list' must be a list or a tuple of strings")
            if len(param_channel_list) != len(det_channels):
                raise RuntimeError(
                    f"Parameter 'param_channel_list' must be 'None' or contain {len(det_channels)} file names."
                )

        channel_num = len(param_channel_list)
        for i in range(channel_num):
            inner_path = "xrfmap/" + det_channel_names[i]
            print(f"Processing data from detector channel {det_channel_names[i]} (#{i+1}) ...")

            # load param file
            param_file_name = param_channel_list[i]

            # load param file
            if not os.path.isabs(param_file_name):
                param_path = os.path.join(working_directory, param_file_name)
            else:
                param_path = param_file_name
            with open(param_path, "r") as json_data:
                param_det = json.load(json_data)

            # update incident energy, required for XANES
            if incident_energy is not None:
                param_det["coherent_sct_energy"]["value"] = incident_energy
                print("Using incident beam energy passed as the function parameter.")
            elif (
                mdata.is_metadata_available()
                and "instrument_mono_incident_energy" in mdata
                and not ignore_datafile_metadata
            ):
                param_det["coherent_sct_energy"]["value"] = mdata["instrument_mono_incident_energy"]
                print(f"Using incident beam energy from the data file '{file_name}'.")
            else:
                print(f"Using incident beam energy from the parameter file '{param_path}'.")

            # The value of incident energy that is used for processing
            incident_energy_used = param_det["coherent_sct_energy"]["value"]
            print(f"Incident beam energy: {incident_energy_used}.")

            data_all_det = data_sets[det_channels[i]].raw_data

            result_map_det, calculation_info = single_pixel_fitting_controller(
                data_all_det,
                param_det,
                incident_energy=incident_energy,
                method=method,
                pixel_bin=pixel_bin,
                raise_bg=raise_bg,
                comp_elastic_combine=comp_elastic_combine,
                linear_bg=linear_bg,
                use_snip=use_snip,
                bin_energy=bin_energy,
                dask_client=dask_client,
            )

            # output to .h5 file
            save_fitdata_to_hdf(fpath, result_map_det, datapath=inner_path)

            def get_scaler_set(img_dict):
                sc_set_names = [_ for _ in img_dict if _.endswith("_scaler")]
                if sc_set_names:
                    return img_dict[sc_set_names[0]]
                else:
                    return {}

            def get_positions_set(img_dict):
                if "positions" in img_dict:
                    return img_dict["positions"]
                else:
                    return {}

            scaler_dict = get_scaler_set(img_dict)
            scaler_name_list = list(scaler_dict.keys())
            positions_dict = get_positions_set(img_dict)
            # Generate dataset
            dataset = copy.deepcopy(scaler_dict)
            dataset.update(result_map_det)

            # Set parameters for quantitative normalization
            param_quant_analysis.experiment_incident_energy = incident_energy_used
            param_quant_analysis.experiment_distance_to_sample = quant_distance_to_sample
            param_quant_analysis.experiment_detector_channel = det_channel_names[i]

            if save_txt is True:
                output_folder = "output_txt_" + prefix_fname
                output_path = os.path.join(working_directory, output_folder)
                output_data(
                    output_dir=output_path,
                    interpolate_to_uniform_grid=interpolate_to_uniform_grid,
                    dataset_name=f"dataset_{det_channel_names[i]}_fit",  # ..._det1_fit, etc.
                    quant_norm=quant_norm,
                    param_quant_analysis=param_quant_analysis,
                    dataset_dict=dataset,
                    positions_dict=positions_dict,
                    file_format="txt",
                    scaler_name=scaler_name,
                    scaler_name_list=scaler_name_list,
                    use_average=use_average,
                )

            if save_tiff is True:
                output_folder = "output_tiff_" + prefix_fname
                output_path = os.path.join(working_directory, output_folder)
                output_data(
                    output_dir=output_path,
                    interpolate_to_uniform_grid=interpolate_to_uniform_grid,
                    dataset_name=f"dataset_{det_channel_names[i]}_fit",  # ..._det1_fit, etc.
                    quant_norm=quant_norm,
                    param_quant_analysis=param_quant_analysis,
                    dataset_dict=dataset,
                    positions_dict=positions_dict,
                    file_format="tiff",
                    scaler_name=scaler_name,
                    scaler_name_list=scaler_name_list,
                    use_average=use_average,
                )

    t1 = time.time()
    print(f"Processing time: {t1 - t0}")


def pyxrf_batch(
    start_id=None,
    end_id=None,
    *,
    param_file_name,
    data_files=None,
    wd=None,
    fit_channel_sum=True,
    fit_channel_each=False,
    param_channel_list=None,
    incident_energy=None,
    ignore_datafile_metadata=False,
    fln_quant_calib_data=None,
    quant_distance_to_sample=0,
    use_snip=True,
    save_txt=False,
    save_tiff=True,
    scaler_name=None,
    use_average=False,
    interpolate_to_uniform_grid=False,
    dask_client=None,
):
    """
    Perform fitting on a batch of data files. The results are saved as new datasets
    in the respective data files and may be viewed using PyXRF. Fitting can be performed on
    the sum of all detector channels, data from selected detector channels, or both.
    Internally, the function is calling the lower level function ``fit_pixel_data_and_save``.
    While it is possible to write processing scripts that call ``fit_pixel_data_and_save`` directly,
    but it is recommended, that ``pyxrf_batch`` is used instead.

    Parameters
    ----------
    start_id : int, optional
        starting run id
    end_id : int, optional
        ending run id
    param_file_name : str, required
        File name of the processing parameter file (JSON) used for data fitting.
        If the parameter file name in the list does not contain full
        path, then it is extended with the path in ``wd``.
    data_files : str, tuple (str,) or list [str], optional
        data file names: may be specified as a string for a single file,
        or iterable container (list, tuple) of strings for multiple files
        If ``data_files`` is specified (not None), then ``start_id`` and ``end_id``
        parameters are ignored. If a file name in the list does not contain full
        path, then it is extended with the path in ``wd``.
    wd : str, or optional
        path folder, default is the current folder
    fit_channel_sum : bool, optional
        fit summed data or not
    fit_channel_each : bool, optional
        fit each channel data or not
    param_channel_list : list, optional
        list of param file names for each channel
    incident_energy : float, optional
        use this energy as incident energy instead of the one in param file, i.e., XANES
        This value overrides the incident energy from metadata and from JSON parameter file.
    ignore_datafile_metadata : bool
        tells whether to ignore metadata from the data file (if data file contains metadata).
        At the moment, only incident energy (monochromator energy) is used by the processing routine.
        If True, then the incident energy from the parameter JSON file is used; if False, then
        the energy from metadata is used. If the function parameter ``incident_energy`` is
        not None, then its value overrides the incident energy from metadata or JSON file.
        Default: False.
    fln_quant_calib_data : str or list(str)
        file name or a list of file names that contain quantitative calibration data
    quant_distance_to_sample : float
        distance-to-sample used in quantitative calibration. If 0, then correction for
        distance is not applied (assumed that the standard and the sample were placed
        at the same distance from the detector).
    use_snip : bool, optional
        use snip method to remove background (`True`). If `False`, then do fitting
        without removing the background (runs faster)
    save_txt : bool, optional
        save data to txt or not
    save_tiff : bool, optional
        save data to tiff or not
    scaler_name : str, optional
        if given, normalization will be performed
    use_average : bool, optional
        if true, norm is performed as data/scaler*mean(scaler), otherwise just data/scaler
    interpolate_to_uniform_grid : bool
        interpolate the result to uniform grid before saving to tiff and txt files
        The grid dimensions match the dimensions of positional data for X and Y axes.
        The range of axes is chosen to fit the values of X and Y.
    dask_client: dask.distributed.Client
        Dask client object. If None, then Dask client is created automatically.
        If a batch of files is processed, then creating Dask client and
        passing the reference to it to the processing functions will save
        execution time:

        .. code:: python

            from pyxrf.api import dask_client_create
            client = dask_client_create()  # Create Dask client
            # <-- code that runs computations -->
            client.close()  # Close Dask client

    Returns
    -------

    flist : [str]
        list of file names of processed files: the list is empty
        if no files were processed. The file names contain relative path
        (combination of ``wd`` and file names).
        The file names are absolute if absolute paths are supplied in ``data_files``.
        If file list is created based on scan IDs, then the paths are always relative.

    Exceptions
    ----------

    ``ValueError`` is raised if the ``param_file_name" is not a string or is referencing
    non-existing file or ``data_files`` is specified, but not a string or iterable object
    containing strings.

    ``IOError`` is raised if data file does not exist. The exception is raised only if
    a single scan is to be processed (``end_id`` is not specied or None) or ``data_files``
    is set to a string with a single data file name. If the range of scan IDs that
    contains one scan (``start_id == end_id``) is specified, or ``data_files`` is
    a list containing a single file name, the exception will not be raised. This
    is consistent with the behavior of the ``make_hdf`` function.

    If an exception is raised by the processing code, it is reraised if
    processing of a single experiment is requested (the same conditions as for ``IOError``).
    Otherwise the error message is printed and processing of the next file or scan in the list
    is attempted.

    The data files and the processing parameter file may be in different directories. If the parameter
    file is in a different directory from the data files, then the full path to the parameter
    file should be specified:

        .. code: python

            param_file_name="/home/user/data_parameters/parameters.json"
            param_file_name="~/data_parameters/parameters.json"

    How does ``pyxrf-batch`` identify a file with a given Scan ID?
    In order for the batch mode processing to work correctly, the data file name must
    conform to the following structure requirements:

        <prefix>_<scanID>(any non-digit character)(any character sequence).h5

    <prefix> - any sequence of characters, not containing ``_``
    <scanID> - string version of Scan ID, obtained as ``str(ScanID)``
    <prefix> and <scanID> are separated by ``_``
    <scanID> is separated from the rest of the file name from the right by any non-number character.
    Files must have extension .h5. (Note, that ``.`` separating the extension is a non-number character
    that may terminate the Scan ID part of the file name

    For example, the following files will be recognized as Scan ID 28355:

        scan2D_28355.h5
        scanNewName_28355_some_comments.h5
        scan_2D_28355.h5  (incorrect, recognized as Scan ID 2, not 28355)
        scan2D28355.h5 (incorrect, no ``_``)

    Examples
    --------

    In the following examples it is assumed that all .h5 files and the .json parameter file are
    located in the current directory.

    -- Process data file with Scan ID 455.

        .. code: python

            pyxrf_batch(455, param_file_name="parameters.json")

    -- Process data files with Scan IDs in the range 455 .. 457:

        .. code: python

            pyxrf_batch(455, 457, param_file_name="parameters.json")

    -- Process all data files in the current directory:

        .. code: python

            pyxrf_batch(param_file_name="parameters.json")

    """

    print()

    if wd is None:
        wd = "."
    else:
        wd = os.path.expanduser(wd)
    param_file_name = os.path.expanduser(param_file_name)

    allow_raising_exceptions = False
    if isinstance(data_files, str):
        allow_raising_exceptions = True
    elif (start_id is not None) and (end_id is None):
        allow_raising_exceptions = True

    if fln_quant_calib_data is not None:
        if not isinstance(fln_quant_calib_data, str) and not isinstance(fln_quant_calib_data, list):
            raise ValueError("Parameter 'fln_quant_calib_data' must be a string or a list of strings")
        if isinstance(fln_quant_calib_data, list) and any([not isinstance(_, str) for _ in fln_quant_calib_data]):
            raise ValueError("List passed with the parameter 'fln_quant_calib_data' must contain only strings")

    if not isinstance(quant_distance_to_sample, float) and not isinstance(quant_distance_to_sample, int):
        raise ValueError("Value of the parameter 'quant_distance_to_sample' must be floating point number")

    if data_files is not None:

        # Check if ``data_files`` has valid value
        data_files_valid = False
        if isinstance(data_files, str):
            data_files = [data_files]  # Convert to list
            data_files_valid = True
        elif isinstance(data_files, Iterable):
            data_files = list(data_files)  # Convert to list
            data_files_valid = True
            for fln in data_files:
                if not isinstance(fln, str):
                    data_files_valid = False
                    break

        if not data_files_valid:
            raise ValueError(
                f"Function pyxrf_batch. Parameter 'data_files' has invalid format. \n"
                f"data_files = '{data_files}'"
            )

        # At this point ``data_files`` is a list of str.
        data_files = [os.path.expanduser(fln) for fln in data_files]

        # Working directory name is appended to each file name in the list,
        data_files = [fln if os.path.isabs(fln) else os.path.join(wd, fln) for fln in data_files]

        # Create the list of files. Include only existing files in the list.
        flist = []
        for fln in data_files:
            if os.path.exists(fln) and os.path.isfile(fln):
                flist.append(fln)
            else:
                if allow_raising_exceptions:
                    raise IOError(f"File '{fln}' does not exist")
                else:
                    print(f"WARNING: file '{fln}' does not exist.")

    else:

        # ``data_files`` parameter is None

        all_files = glob.glob(os.path.join(wd, "*.h5"))

        # Sort files (processing unsorted list of files is unsightly)
        all_files.sort()

        if start_id is None and end_id is None:
            # ``start_id`` and ``end_id`` are not specified:
            #   process all .h5 files in the current directory
            flist = all_files
        elif end_id is None:
            # only ``start_id`` is specified:
            #   process only one file that contains ``start_id`` in its name
            #   (only if such file exists)
            pattern = f"^[^_]*_{str(start_id)}\D+"  # noqa: W605
            flist = [fname for fname in all_files if re.search(pattern, os.path.basename(fname))]

            if len(flist) < 1:
                msg = f"File with Scan ID {start_id} was not found"
                if allow_raising_exceptions:
                    raise IOError(msg)
                else:
                    print(msg)
            else:
                print(f"Processing file with Scan ID {start_id}")
        else:
            # ``start_id`` and ``end_id`` are specified:
            #   select files, which contain the respective ID substring in their names
            flist = []
            for data_id in range(start_id, end_id + 1):
                pattern = f"^[^_]*_{str(data_id)}\D+"  # noqa: W605
                flist += [fname for fname in all_files if re.search(pattern, os.path.basename(fname))]
            if len(flist) < 1:
                print(f"No files with Scan IDs in the range {start_id} .. {end_id} were found.")
            else:
                print(f"Processing file with Scan IDs in the range {start_id} .. {end_id}")

    # Check if .json parameter file exists
    if not os.path.isabs(param_file_name):
        pname = os.path.join(wd, param_file_name)
    else:
        pname = param_file_name
    if not os.path.exists(pname) or not os.path.isfile(pname):
        raise IOError(f"Function 'pyxrf_batch'. Processing parameter file '{pname}' does not exist.")
    else:
        print(f"Processing parameter file: '{pname}'")

    if len(flist) > 0:
        # If no external Dask client is provided and we are processing a batch
        #   then create a local client that will be used to process the whole batch
        if (len(flist) > 1) and (dask_client is None):
            logger.info("Creating local Dask client for processing the batch of files ...")
            dask_client = dask_client_create()
            client_is_local = True
        else:
            client_is_local = False

        def _dask_client_close(is_local):
            # We don't want to close an externally provided client
            if is_local:
                logger.info("Closing the local Dask client ...")
                dask_client.close()

        print("The following files are scheduled for processing:")
        for fln in flist:
            print(f"    {fln}")
        print(f"Total number of selected files: {len(flist)}\n")

        for fpath in flist:
            print(f"Processing file '{fpath}' ...")
            fname = fpath.split(sep_v)[-1]
            working_directory = fpath[: -len(fname)]
            try:
                fit_pixel_data_and_save(
                    working_directory,
                    fname,
                    fit_channel_sum=fit_channel_sum,
                    param_file_name=param_file_name,
                    fit_channel_each=fit_channel_each,
                    param_channel_list=param_channel_list,
                    incident_energy=incident_energy,
                    ignore_datafile_metadata=ignore_datafile_metadata,
                    fln_quant_calib_data=fln_quant_calib_data,
                    quant_distance_to_sample=quant_distance_to_sample,
                    use_snip=use_snip,
                    save_txt=save_txt,
                    save_tiff=save_tiff,
                    scaler_name=scaler_name,
                    use_average=use_average,
                    interpolate_to_uniform_grid=interpolate_to_uniform_grid,
                    dask_client=dask_client,
                )
            except Exception as ex:
                if allow_raising_exceptions:
                    _dask_client_close(client_is_local)
                    raise Exception from ex
                else:
                    print(f"ERROR: could not process the file '{fname}'. No results are saved.")

        print("\nAll selected files were processed.")

        _dask_client_close(client_is_local)

    else:

        print("No files were selected for processing.")

    return flist


def fit_each_pixel_with_nnls(data, params, elemental_lines=None, incident_energy=None, weights=None):
    """
    Fit a spectrum with a linear model.

    Parameters
    ----------
    data : array
        spectrum intensity
    param : dict
        fitting parameters
    elemental_lines : list, optional
            e.g., ['Na_K', Mg_K', 'Pt_M'] refers to the
            K lines of Sodium, the K lines of Magnesium, and the M
            lines of Platinum. If elemental_lines is set as None,
            all the possible lines activated at given energy will be used.
    """
    param = copy.deepcopy(params)
    if incident_energy is not None:
        param["coherent_sct_amplitude"]["value"] = incident_energy
    # cut data into proper range
    low = param["non_fitting_values"]["energy_bound_low"]["value"]
    high = param["non_fitting_values"]["energy_bound_high"]["value"]
    a0 = param["e_offset"]["value"]
    a1 = param["e_linear"]["value"]
    x, y = define_range(data, low, high, a0, a1)
    # pixel fitting
    _, result_dict, area_dict = linear_spectrum_fitting(x, y, elemental_lines=elemental_lines, weights=weights)
    return result_dict


def fit_pixel_per_file_no_multi(dir_path, file_prefix, fileID, param, interpath, save_spectrum=True):
    """
    Single pixel fit of experiment data. No multiprocess is applied.

    .. warning :: This function is not optimized as it calls
    linear_spectrum_fitting function, where lots of repeated
    calculation are processed.

    Parameters
    ----------
    data : array
        3D data of experiment spectrum
    param : dict
        fitting parameters

    Returns
    -------
    dict :
        fitting values for all the elements
    """

    num_str = "{:03d}".format(fileID)
    filename = file_prefix + num_str
    file_path = os.path.join(dir_path, filename)
    with h5py.File(file_path, "r") as f:
        data = f[interpath][:]
    datas = data.shape

    elist = param["non_fitting_values"]["element_list"].split(", ")
    elist = [e.strip(" ") for e in elist]

    non_element = ["compton", "elastic", "background"]
    total_list = elist + non_element

    result_map = dict()
    for v in total_list:
        if save_spectrum:
            result_map.update({v: np.zeros([datas[0], datas[1], datas[2]])})
        else:
            result_map.update({v: np.zeros([datas[0], datas[1]])})

    for i in range(datas[0]):
        for j in range(datas[1]):
            x, result, area_v = linear_spectrum_fitting(
                data[i, j, :], param, elemental_lines=elist, constant_weight=1.0
            )
            for v in total_list:
                if v in result:
                    if save_spectrum:
                        result_map[v][i, j, : len(result[v])] = result[v]
                    else:
                        result_map[v][i, j] = np.sum(result[v])

    return result_map


def fit_data_multi_files(dir_path, file_prefix, param, start_i, end_i, interpath="entry/instrument/detector/data"):
    """
    Fitting for multiple files with Multiprocessing.

    Parameters
    ----------
    dir_path : str
    file_prefix : str
    param : dict
    start_i : int
        start id of given file
    end_i: int
        end id of given file
    interpath : str
        path inside hdf5 file to fetch the data

    Returns
    -------
    result : list
        fitting result as list of dict
    """
    num_processors_to_use = multiprocessing.cpu_count()
    logger.info("cpu count: {}".format(num_processors_to_use))
    pool = multiprocessing.Pool(num_processors_to_use)

    result_pool = [
        pool.apply_async(fit_pixel_per_file_no_multi, (dir_path, file_prefix, m, param, interpath))
        for m in range(start_i, end_i + 1)
    ]

    results = []
    for r in result_pool:
        results.append(r.get())

    pool.terminate()
    pool.join()
    return results


if __name__ == "__main__":
    logger = logging.getLogger(__name__)

    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(fmt="%(asctime)s : %(levelname)s : %(message)s")

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)

    # pyxrf_batch(start_id=92276, end_id=92276,
    #             param_file_name="param_335",
    #             wd=".", save_tiff=False)

    pyxrf_batch(
        start_id=63339,
        param_file_name="test",
        wd=".",
        save_tiff=True,
        fln_quant_calib_data=["standard_41147a.json"],
        quant_distance_to_sample=2.0,
        scaler_name="i0",
        # fln_quant_calib_data=['standard_32654A.json',
        #                       'standard_32654A a.json',
        #                       'standard_41147.json'],
        fit_channel_sum=True,
        fit_channel_each=True,
        # param_channel_list=['det1', 'det2', 'det3']
    )



import re
import xraylib
from distutils.version import LooseVersion

from skbeam.core.constants import XrfElement as Element
from skbeam.core.fitting.xrf_model import K_LINE, L_LINE, M_LINE


def get_element_atomic_number(element_str):
    r"""
    A wrapper to ``SymbolToAtomicNumber`` function from ``xraylib``.
    Returns atomic number for the sybmolic element name (e.g. ``C`` or ``Fe``).

    Parameters
    ----------

    element_str: str
        sybmolic representation of an element

    Returns
    -------

    Atomic number of the element ``element_str``. If element is invalid, then
    the function returns 0.

    """
    if LooseVersion(xraylib.__version__) < LooseVersion("4.0.0"):
        xraylib.SetErrorMessages(0)  # Turn off error messages from ``xraylib``

    try:
        val = xraylib.SymbolToAtomicNumber(element_str)
    except ValueError:
        # Imitate the behavior of xraylib 3
        val = 0
    return val


def validate_element_str(element_str):
    r"""
    Checks if ``element_str`` is valid representation of an element according to
    standard notation for chemical formulas. Valid representation of elements can
    be processed by ``xraylib`` tools. This functions attempts to find the atomic
    number for the element and returns ``True`` if it succeeds and ``False`` if
    it fails.

    Parameters
    ----------

    element_str: str
        sybmolic representation of an element

    Returns
    -------

    ``True`` if ``element_str`` is valid representation of an element and ``False``
    otherwise.
    """

    if get_element_atomic_number(element_str):
        return True
    else:
        return False


def parse_compound_formula(compound_formula):
    r"""
    Parses the chemical formula of a compound and returns the dictionary,
    which contains element name, atomic number, number of atoms and mass fraction
    in the compound.

    Parameters
    ----------

    compound_formula: str
        chemical formula of the compound in the form ``FeO2``, ``CO2`` or ``Fe``.
        Element names must start with capital letter.

    Returns
    -------

        dictionary of dictionaries, data on each element in the compound: key -
        sybolic element name, value - a dictionary that contains ``AtomicNumber``,
        ``nAtoms`` and ``massFraction`` of the element. The elements are sorted
        in the order of growing atomic number.

    Raises
    ------

        RuntimeError is raised if compound formula cannot be parsed
    """

    if LooseVersion(xraylib.__version__) < LooseVersion("4.0.0"):
        xraylib.SetErrorMessages(0)  # This is supposed to stop XRayLib from printing
        #                              internal error messages, but it doesn't work

    try:
        compound_data = xraylib.CompoundParser(compound_formula)
    except (SystemError, ValueError):
        msg = f"Invalid chemical formula '{compound_formula}' is passed, parsing failed"
        raise RuntimeError(msg)

    # Now create more manageable structure
    compound_dict = {}
    for e_an, e_mf, e_na in zip(
        compound_data["Elements"], compound_data["massFractions"], compound_data["nAtoms"]
    ):
        e_name = xraylib.AtomicNumberToSymbol(e_an)
        compound_dict[e_name] = {"AtomicNumber": e_an, "nAtoms": e_na, "massFraction": e_mf}

    return compound_dict


def split_compound_mass(compound_formula, compound_mass):
    r"""
    Computes mass of each element in the compound given total mass of the compound

    Parameters
    ----------

    compound_formula: str
        chemical formula of the compound in the form ``FeO2``, ``CO2`` or ``Fe``.
        Element names must start with capital letter.

    compound_mass: float
        total mass of the compound

    Returns
    -------

        dictionary: key - symbolic element name, value - mass of the element

    Raises
    ------

        RuntimeError is raised if compound formula cannot be parsed
    """

    compound_dict = parse_compound_formula(compound_formula)

    element_dict = {}
    for el_name, el_info in compound_dict.items():
        element_dict[el_name] = el_info["massFraction"] * compound_mass

    return element_dict


def get_supported_eline_list(*, lines=None):
    """
    Returns the list of the emission lines supported by ``scikit-beam``

    Parameters
    ----------

    lines : list(str)
        tuple or list of strings, that defines, which emission lines are going to be included
        in the output list (e.g. ``("K",)`` or ``("L", "M")`` etc.) If ``None`` (default),
        then K, L and M lines are going to be included.

    Returns
    -------
        the list of supported emission line. The lines are in the format ``"Fe_K"`` or ``"Mg_M"``.
    """

    if lines is None:
        lines = ("K", "L", "M")

    eline_list = []
    if "K" in lines:
        eline_list += K_LINE
    if "L" in lines:
        eline_list += L_LINE
    if "M" in lines:
        eline_list += M_LINE

    return eline_list


def check_if_eline_supported(eline_name, *, lines=None):
    """
    Check if the emission line name is in the list of supported names.
    Emission name must be in the format: K_K, Fe_K etc. The list includes K, L and M lines.
    The function is case-sensitive.

    Parameters
    ----------
    eline_name : str
        name of the emission line (K_K, Fe_K etc. for valid emission line). In general
        the string may contain arbitrary sequence characters, may be empty or None. The
        function will return True only if the sequence represents emission line from
        the list of supported emission lines.

    lines : list(str)
        tuple or list of strings, that defines, which emission lines are going to be included
        in the output list (e.g. ``("K",)`` or ``("L", "M")`` etc.) If ``None`` (default),
        then K, L and M lines are going to be included.

    Returns
        True if ``eline_name`` is in the list of supported emission lines, False otherwise
    """
    if not eline_name or not isinstance(eline_name, str):
        return False

    if eline_name in get_supported_eline_list(lines=lines):
        return True
    else:
        return False


def check_if_eline_is_activated(elemental_line, incident_energy):
    """
    Checks if emission line is activated at given incident beam energy

    Parameters
    ----------

    elemental_line : str
        emission line in the format K_K or Fe_K
    incident_energy : float
        incident energy in keV

    Returns
    -------
        bool value, indicating if the emission line is activated
    """

    # Check if the emission line has correct format
    if not re.search(r"^[A-Z][a-z]?_[KLM]([ab]\d?)?$", elemental_line):
        raise RuntimeError(f"Elemental line {elemental_line} is improperly formatted")

    # The validation of 'elemental_line' is strict enough to do the rest of the processing
    #   without further checks.
    [element, line] = elemental_line.split("_")
    line = line.lower()
    if len(line) == 1:
        line += "a1"
    elif len(line) == 2:
        line += "1"

    e = Element(element)
    if e.cs(incident_energy)[line] == 0:
        return False
    else:
        return True


def generate_eline_list(element_list, *, incident_energy, lines=None):
    r"""
    Generate a list of emission lines based on the list of elements (``element_list``)
    and incident energy. Only the emission lines that are supported by ``scikit-beam``
    and activated by the incident energy are included in the list.

    Parameters
    ----------

    element_list: list(str)
        list of valid element names (e.g. ["S", "Al", "Fe"])

    incident_energy: float
        incident beam energy, keV

    lines: list(str)
        tuple or list of strings, that defines, which classes of emission lines to include
        in the output list (e.g. ``("K",)`` or ``("L", "M")`` etc.) If ``None`` (default),
        then K, L and M lines are going to be included.

    Returns
    -------

    list(str) - the list of emission lines

    Raises
    ------

    RuntimeError is raised if 'lines' contains incorrect specification of emission line class.
    """

    if lines is None:
        # By default lines "K", "L" and "M" are included in the output list
        lines = ("K", "L", "M")

    # Verify line selection
    invalid_lines = []
    for ln in lines:
        if not re.search(r"^[KLM]$", ln):
            invalid_lines.append(ln)
    if invalid_lines:
        msg = f"Some of the selected emission lines are incorrect: {invalid_lines}"
        raise RuntimeError(msg)

    eline_list = []

    for element in element_list:
        for ln in lines:
            eline = f"{element}_{ln}"
            is_activated = check_if_eline_is_activated(eline, incident_energy)
            is_supported = check_if_eline_supported(eline)
            if is_activated and is_supported:
                eline_list.append(eline)

    return eline_list


# TODO: the following function needs tests
def get_eline_parameters(elemental_line, incident_energy):
    """
    Returns emission line parameters

    Parameters
    ----------

    elemental_line : str
        emission line in the format K_K, Fe_K, Ca_k, Ca_ka, Ca_kb2 etc.
    incident_energy : float
        incident energy in keV

    Returns
    -------
    dict
        Computed parameters of the emission line. Keys: "energy" (central energy of the peak),
        "cs" (crossection), "ratio" (normalized crossection, e.g. cs(Ca_kb2)/ cs(Ca_ka1)
    """

    # Check if the emission line has correct format
    # TODO: verify the correnct range for lines (a-z covers all the cases, but may be too broad)
    if not re.search(r"^[A-Z][a-z]?_[KLMklm]([a-z]\d?)?$", elemental_line):
        raise RuntimeError(f"Elemental line {elemental_line} is improperly formatted")

    # The validation of 'elemental_line' is strict enough to do the rest of the processing
    #   without further checks.
    [element, line] = elemental_line.split("_")
    line = line.lower()
    if len(line) == 1:
        line += "a1"
    elif len(line) == 2:
        line += "1"

    # This is the name of line #1 (ka1, la1 etc.)
    line_1 = line[0] + "a1"

    try:
        e = Element(element)
        energy = e.emission_line[line]
        cs = e.cs(incident_energy)[line]
        cs_1 = e.cs(incident_energy)[line_1]
        ratio = cs / cs_1 if cs_1 else 0
    except Exception:
        energy, cs, ratio = 0, 0, 0

    return {"energy": energy, "cs": cs, "ratio": ratio}


# TODO: the following function needs tests
def get_eline_energy(elemental_line):
    """
    Returns emission line parameters

    Parameters
    ----------

    elemental_line : str
        emission line in the format K_K, Fe_K, Ca_k, Ca_ka, Ca_kb2 etc.

    Returns
    -------
    float
        Peak center energy for the emission line, keV.
    """

    # Check if the emission line has correct format
    # TODO: verify the correnct range for lines (a-z covers all the cases, but may be too broad)
    if not re.search(r"^[A-Z][a-z]?_[KLMklm]([a-z]\d?)?$", elemental_line):
        raise RuntimeError(f"Elemental line {elemental_line} is improperly formatted")

    # The validation of 'elemental_line' is strict enough to do the rest of the processing
    #   without further checks.
    [element, line] = elemental_line.split("_")
    line = line.lower()
    if len(line) == 1:
        line += "a1"
    elif len(line) == 2:
        line += "1"

    try:
        e = Element(element)
        energy = e.emission_line[line]
    except Exception:
        energy = 0

    return energy



import copy
import pytest

from pyxrf.gui_module.useful_widgets import global_gui_variables
from pyxrf.gui_module.wnd_image_wizard import WndImageWizard

# =====================================================================
#                 class WndImageWizard (SecondaryWindow)

# fmt: off
sample_table_content = [
    ["Ar_K", 4.277408, 307.452117],
    ["Ca_K", 0.000000, 1.750295e+03],
    ["Fe_K", 17.902211, 4.576803e+04],
    ["Tb_L", 0.000000, 2.785734e+03],
    ["Ti_K", 0.055362, 1.227637e+04],
    ["Zn_K", 0.000000, 892.008670],
    ["compton", 0.000000, 249.055352],
    ["elastic", 0.000000, 163.153881],
    ["i0", 1.715700e+04, 1.187463e+06],
    ["i0_time", 3.066255e+06, 1.727313e+08],
]
# fmt: on


def _get_gpc_sim(maps_info_table):
    """Simulated simplified 'GlobalProcessingClasses` class for testing"""

    class SimGlobalProcessingClasses:
        def __init__(self):
            self.range_table = copy.deepcopy(maps_info_table)
            # Selection limit table may be the same
            self.limit_table = copy.deepcopy(maps_info_table)
            # Show status table
            self.show_table = [[_[0], False] for _ in self.range_table]

        def get_maps_info_table(self):
            # TODO: implementation needed
            return self.range_table, self.limit_table, self.show_table

    return SimGlobalProcessingClasses()


# fmt: off
@pytest.mark.parametrize("sample_table_data", [
    [],  # Empty table
    [sample_table_content[0]],  # Only one table row
    sample_table_content])  # Full table
# fmt: on
def test_WndImageWizard_1(qtbot, sample_table_data):
    """
    Fill the table and show the window.
    Test if the table data is set successfully and the table contains
    the expected number of rows.
    """
    gpc = _get_gpc_sim(sample_table_data)
    gui_vars = global_gui_variables

    wnd = WndImageWizard(gpc=gpc, gui_vars=gui_vars)
    qtbot.addWidget(wnd)
    wnd.show()

    wnd.slot_update_table()
    # Check that the table was saved correctly
    assert wnd._range_data == sample_table_data, "Table data was set incorrectly"
    # Check the number of table rows
    assert wnd.table.rowCount() == len(sample_table_data), "Incorrect number of table rows"



import textwrap

from qtpy.QtWidgets import (
    QVBoxLayout,
    QLabel,
    QDialog,
    QDialogButtonBox,
    QTableWidget,
    QTableWidgetSelectionRange,
    QTableWidgetItem,
    QHeaderView,
)
from qtpy.QtGui import QBrush, QColor
from qtpy.QtCore import Qt

from .useful_widgets import set_tooltip

"""
# This table may be useful for unit tests that are yet to be written. So keep it for now.
sample_table_contents = [
    ["41147", "Micromatter 41147",
     "GaP 21.2 (Ga=15.4, P=5.8) / CaF2 14.6 / V 26.4 / Mn 17.5 / Co 21.7 / Cu 20.3"],
    ["41148", "Micromatter 41148",
     "GaP 21.1 (Ga=15.4, P=5.7) / CaF2 14.6 / V 25.1 / Mn 18.1 / Co 20.6 / Cu 20.5"],
    ["41151", "Micromatter 41151",
     "KCl 21.9 / Ti 21.3 / Fe 24.4 / ZnTe 20.3 / Pb 22.3"],
    ["41164", "Micromatter 41164",
     "CeF3 21.1 / Au 20.6"],
]
"""


class DialogSelectQuantStandard(QDialog):
    def __init__(self, parent=None):

        super().__init__(parent)

        self._qe_param_built_in = []
        self._qe_param_custom = []
        self._qe_standard_selected = None
        self._qe_param = []  # The list of all standards
        self._custom_label = []  # The list of booleans: True - the standard is custom, False -built-in

        self.setWindowTitle("Load Quantitative Standard")

        self.setMinimumHeight(500)
        self.setMinimumWidth(600)
        self.resize(600, 500)

        self.selected_standard_index = -1

        labels = ("C", "Serial #", "Name", "Description")
        col_stretch = (
            QHeaderView.ResizeToContents,
            QHeaderView.ResizeToContents,
            QHeaderView.ResizeToContents,
            QHeaderView.Stretch,
        )
        self.table = QTableWidget()
        set_tooltip(
            self.table,
            "To <b> load the sta"
            "ndard</b>, double-click on the table row or "
            "select the table row and then click <b>Ok</b> button.",
        )
        self.table.setMinimumHeight(200)
        self.table.setColumnCount(len(labels))
        self.table.verticalHeader().hide()
        self.table.setHorizontalHeaderLabels(labels)
        self.table.setSelectionBehavior(QTableWidget.SelectRows)
        self.table.setSelectionMode(QTableWidget.SingleSelection)
        self.table.itemSelectionChanged.connect(self.item_selection_changed)
        self.table.itemDoubleClicked.connect(self.item_double_clicked)

        self.table.setStyleSheet(
            "QTableWidget::item{color: black;}"
            "QTableWidget::item:selected{background-color: red;}"
            "QTableWidget::item:selected{color: white;}"
        )

        header = self.table.horizontalHeader()
        for n, col_stretch in enumerate(col_stretch):
            # Set stretching for the columns
            header.setSectionResizeMode(n, col_stretch)

        self.lb_info = QLabel()
        self.lb_info.setText("Column 'C': * means that the standard is user-defined.")

        button_box = QDialogButtonBox(QDialogButtonBox.Open | QDialogButtonBox.Cancel)
        button_box.button(QDialogButtonBox.Cancel).setDefault(True)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        self.pb_open = button_box.button(QDialogButtonBox.Open)
        self.pb_open.setEnabled(False)

        vbox = QVBoxLayout()
        vbox.addWidget(self.table)
        vbox.addWidget(self.lb_info)
        vbox.addWidget(button_box)
        self.setLayout(vbox)

    def _fill_table(self, table_contents):
        self.table.setRowCount(len(table_contents))

        for nr, row in enumerate(table_contents):
            for nc, entry in enumerate(row):
                s = textwrap.fill(entry, width=40)
                item = QTableWidgetItem(s)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                if not nc:
                    item.setTextAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
                self.table.setItem(nr, nc, item)
        self.table.resizeRowsToContents()

        brightness = 220
        for nr in range(self.table.rowCount()):
            for nc in range(self.table.columnCount()):
                self.table.item(nr, nc)
                if nr % 2:
                    color = QColor(255, brightness, brightness)
                else:
                    color = QColor(brightness, 255, brightness)
                self.table.item(nr, nc).setBackground(QBrush(color))

        try:
            index = self._qe_param.index(self._qe_standard_selected)
            self.selected_standard_index = index
            n_columns = self.table.columnCount()
            self.table.setRangeSelected(QTableWidgetSelectionRange(index, 0, index, n_columns - 1), True)
        except ValueError:
            pass

    def item_selection_changed(self):
        sel_ranges = self.table.selectedRanges()
        # The table is configured to have one or no selected ranges
        # 'Open' button should be enabled only if a range (row) is selected
        if sel_ranges:
            self.selected_standard_index = sel_ranges[0].topRow()
            self.pb_open.setEnabled(True)
        else:
            self.selected_standard_index = -1
            self.pb_open.setEnabled(False)

    def item_double_clicked(self):
        self.accept()

    def set_standards(self, qe_param_built_in, qe_param_custom, qe_standard_selected):
        self._qe_standard_selected = qe_standard_selected
        self._qe_param = qe_param_custom + qe_param_built_in
        custom_label = [True] * len(qe_param_custom) + [False] * len(qe_param_built_in)

        table_contents = []
        for n, param in enumerate(self._qe_param):
            custom = "*" if custom_label[n] else ""
            serial = param["serial"]
            name = param["name"]
            description = param["description"]
            table_contents.append([custom, serial, name, description])

        self._fill_table(table_contents)

    def get_selected_standard(self):
        if self.selected_standard_index >= 0:
            return self._qe_param[self.selected_standard_index]
        else:
            return None



import os

from qtpy.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QCheckBox,
    QLabel,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QTextEdit,
)
from qtpy.QtGui import QDoubleValidator
from qtpy.QtCore import Qt

from .useful_widgets import (
    LineEditReadOnly,
    PushButtonMinimumWidth,
    set_tooltip,
    LineEditExtended,
    DoubleValidatorStrict,
)

import logging

logger = logging.getLogger(__name__)


class DialogSaveCalibration(QDialog):
    def __init__(self, parent=None, *, file_path=None):

        super().__init__(parent)

        self.__file_path = ""
        self.__distance_to_sample = 0.0
        self.__overwrite_existing = False
        self.__preview = ("", {})  # str - information, dict - warnings

        self.setWindowTitle("Save Quantitative Calibration")
        self.setMinimumHeight(600)
        self.setMinimumWidth(600)
        self.resize(600, 600)

        self.text_edit = QTextEdit()
        set_tooltip(
            self.text_edit,
            "Preview the <b>quantitative calibration data</b> to be saved. The displayed "
            "warnings will not be saved, but need to be addressed in order to keep "
            "data integrity. The parameter <b>distance-to-sample</b> is optional, "
            "but desirable. If <b>distance-to-sample</b> is zero then no scaling will be "
            "applied to data to compensate for changing distance.",
        )
        self.text_edit.setReadOnly(True)

        self.le_file_path = LineEditReadOnly()
        set_tooltip(
            self.le_file_path,
            "Full path to the file used to <b>save the calibration data</b>. The path "
            "can be changed in file selection dialog box.",
        )
        self.pb_file_path = PushButtonMinimumWidth("..")
        set_tooltip(self.pb_file_path, "Change <b>file path</b> for saving the calibration data.")
        self.pb_file_path.clicked.connect(self.pb_file_path_clicked)
        self.pb_file_path.setDefault(False)
        self.pb_file_path.setAutoDefault(False)

        self.le_distance_to_sample = LineEditExtended()
        self.le_distance_to_sample.textChanged.connect(self.le_distance_to_sample_text_changed)
        self.le_distance_to_sample.editingFinished.connect(self.le_distance_to_sample_editing_finished)
        self._le_distance_to_sample_validator = DoubleValidatorStrict()
        self._le_distance_to_sample_validator.setBottom(0.0)
        set_tooltip(
            self.le_distance_to_sample,
            "<b>Distance</b> between the detector and the sample during calibration. If the value "
            "is 0, then no scaling is applied to data to correct the data if distance-to-sample "
            "is changed between calibration and measurement.",
        )

        self.cb_overwrite = QCheckBox("Overwrite Existing")
        self.cb_overwrite.stateChanged.connect(self.cb_overwrite_state_changed)
        set_tooltip(
            self.cb_overwrite,
            "Overwrite the <b>existing</b> file. This is a safety feature implemented to protect "
            "valuable results from accidental deletion.",
        )

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("The following data will be saved to JSON file:"))
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addWidget(self.text_edit)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Path: "))
        hbox.addWidget(self.pb_file_path)
        hbox.addWidget(self.le_file_path)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Distance-to-sample:"))
        hbox.addWidget(self.le_distance_to_sample)
        hbox.addStretch(1)
        hbox.addWidget(self.cb_overwrite)
        vbox.addLayout(hbox)

        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.pb_ok = button_box.button(QDialogButtonBox.Ok)
        self.pb_ok.setDefault(False)
        self.pb_ok.setAutoDefault(False)
        self.pb_cancel = button_box.button(QDialogButtonBox.Cancel)
        self.pb_cancel.setDefault(True)
        self.pb_cancel.setAutoDefault(True)

        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        vbox.addWidget(button_box)

        self.setLayout(vbox)

        self._show_distance_to_sample()
        self._show_preview()
        self._show_overwrite_existing()

        # Set and display file path
        if file_path is not None:
            self.file_path = file_path

    @property
    def file_path(self):
        return self.__file_path

    @file_path.setter
    def file_path(self, file_path):
        file_path = os.path.expanduser(file_path)
        self.__file_path = file_path
        self.le_file_path.setText(file_path)

    @property
    def distance_to_sample(self):
        return self.__distance_to_sample

    @distance_to_sample.setter
    def distance_to_sample(self, distance_to_sample):
        self.__distance_to_sample = distance_to_sample
        self._show_distance_to_sample()

    @property
    def overwrite_existing(self):
        return self.__overwrite_existing

    @overwrite_existing.setter
    def overwrite_existing(self, overwrite_existing):
        self.__overwrite_existing = overwrite_existing
        self._show_overwrite_existing()

    @property
    def preview(self):
        return self.__preview

    @preview.setter
    def preview(self, preview):
        self.__preview = preview
        self._show_preview()

    def pb_file_path_clicked(self):
        file_path = QFileDialog.getSaveFileName(
            self,
            "Select File to Save Quantitative Calibration",
            self.file_path,
            "JSON (*.json);; All (*)",
            options=QFileDialog.DontConfirmOverwrite,
        )
        file_path = file_path[0]
        if file_path:
            self.file_path = file_path

    def le_distance_to_sample_text_changed(self, text):
        valid = self._le_distance_to_sample_validator.validate(text, 0)[0] == QDoubleValidator.Acceptable
        self.le_distance_to_sample.setValid(valid)
        self.pb_ok.setEnabled(valid)

    def le_distance_to_sample_editing_finished(self):
        text = self.le_distance_to_sample.text()
        if self._le_distance_to_sample_validator.validate(text, 0)[0] == QDoubleValidator.Acceptable:
            self.__distance_to_sample = float(text)
        self._show_distance_to_sample()
        self._show_preview()  # Show/hide warning on zero distance-to-sample value

    def cb_overwrite_state_changed(self, state):
        state = state == Qt.Checked
        self.__overwrite_existing = state

    def _show_distance_to_sample(self):
        self.le_distance_to_sample.setText(f"{self.__distance_to_sample:.10g}")

    def _show_overwrite_existing(self):
        self.cb_overwrite.setChecked(Qt.Checked if self.__overwrite_existing else Qt.Unchecked)

    def _show_preview(self):
        text = ""
        # First print warnings
        for key, value in self.__preview[1].items():
            if "distance" in key and self.__distance_to_sample > 0:
                continue
            text += value + "\n"
        # Additional space if there are any warinings
        if len(text):
            text += "\n"
        # Then add the main block of text
        text += self.__preview[0]
        self.text_edit.setText(text)



from qtpy.QtWidgets import QFrame, QSplitter, QHBoxLayout
from qtpy.QtCore import Qt
from .left_panel import LeftPanel
from .right_panel import RightPanel


class TwoPanelWidget(QSplitter):
    def __init__(self, *, gpc, gui_vars):

        super().__init__(Qt.Horizontal)

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.frame_left = QFrame(self)
        self.frame_left.setFrameShape(QFrame.StyledPanel)

        self.frame_right = QFrame(self)
        self.frame_right.setFrameShape(QFrame.StyledPanel)

        self.addWidget(self.frame_left)
        self.addWidget(self.frame_right)

        hbox = QHBoxLayout()
        self.left_panel = LeftPanel(gpc=self.gpc, gui_vars=self.gui_vars)
        hbox.addWidget(self.left_panel)
        self.frame_left.setLayout(hbox)

        hbox = QHBoxLayout()
        self.right_panel = RightPanel(gpc=self.gpc, gui_vars=self.gui_vars)
        hbox.addWidget(self.right_panel)
        self.frame_right.setLayout(hbox)

        self._show_first_time = True

        # Set stretch factor of the left panel to 0 (we want it to keep its width
        #   as the window is resized. The panel can still be resized manually
        self.setStretchFactor(0, 0)
        # Set stretch factor for the right panel to some non-zero value, e.g. 1
        self.setStretchFactor(1, 1)

    def showEvent(self, event):
        # Set the ratio for the splitter (only the first time the window is shown)
        if self._show_first_time:
            self.setSizes([460, self.width() - 460])
            self._show_first_time = False

    def update_widget_state(self, condition=None):
        self.left_panel.update_widget_state(condition)
        self.right_panel.update_widget_state(condition)



import json
import yaml

# =========================================================================
#   Data for testing WndLoadQuantitativeCalibration class

# Two sets of quantitative calibration data for demonstration of GUI layout
#   Real data will be used when GUI is integrated with the program.
#   The data are representing contents of JSON files, so they should be loaded
#   using 'json' module.
json_quant_calib_1 = """
    {
        "name": "Micromatter 41147",
        "serial": "v41147",
        "description": "GaP 21.2 (Ga=15.4, P=5.8) / CaF2 14.6 / V 26.4 / Mn 17.5 / Co 21.7 / Cu 20.3",
        "element_lines": {
            "Ga_K": {
                "density": 15.4,
                "fluorescence": 6.12047035678267e-05
            },
            "Ga_L": {
                "density": 15.4,
                "fluorescence": 1.1429814846741588e-05
            },
            "P_K": {
                "density": 5.8,
                "fluorescence": 3.177988019213722e-05
            },
            "F_K": {
                "density": 7.105532786885246,
                "fluorescence": 1.8688801284649113e-07
            },
            "Ca_K": {
                "density": 7.494467213114753,
                "fluorescence": 0.0005815345261894806
            },
            "V_K": {
                "density": 26.4,
                "fluorescence": 0.00030309931019669974
            },
            "Mn_K": {
                "density": 17.5,
                "fluorescence": 0.0018328847495676865
            },
            "Co_K": {
                "density": 21.7,
                "fluorescence": 0.0014660067400157218
            },
            "Cu_K": {
                "density": 20.3,
                "fluorescence": 6.435121428993609e-05
            }
        },
        "incident_energy": 12.0,
        "detector_channel": "sum",
        "scaler_name": "i0",
        "distance_to_sample": 1.0,
        "creation_time_local": "2020-05-27T18:49:14+00:00",
        "source_scan_id": null,
        "source_scan_uid": null
    }
    """

json_quant_calib_2 = """
    {
        "name": "Micromatter 41164 Name Is Long So It Has To Be Printed On Multiple Lines (Some More Words To Make The Name Longer)",
        "serial": "41164",
        "description": "CeF3 21.1 / Au 20.6",
        "element_lines": {
            "F_K": {
                "density": 6.101050068482728,
                "fluorescence": 2.1573457185882552e-07
            },
            "Ce_L": {
                "density": 14.998949931517274,
                "fluorescence": 0.0014368335445700924
            },
            "Au_L": {
                "density": 20.6,
                "fluorescence": 4.4655757003090785e-05
            },
            "Au_M": {
                "density": 20.6,
                "fluorescence": 3.611978659032483e-05
            }
        },
        "incident_energy": 12.0,
        "detector_channel": "sum",
        "scaler_name": "i0",
        "distance_to_sample": 2.0,
        "creation_time_local": "2020-05-27T18:49:53+00:00",
        "source_scan_id": null,
        "source_scan_uid": null
    }
    """  # noqa: E501

# The data is structured the same way as in the actual program code, so transitioning
#   to real data will be simple
quant_calib = [
    [json.loads(json_quant_calib_1), {"file_path": "/path/to/quantitative/calibration/file/standard_41147.json"}],
    [
        json.loads(json_quant_calib_2),
        {
            "file_path": "/extremely/long/path/to"
            "/quantitative/calibration/file/so/it/had/to/be/"
            "printed/on/multiple/lines/standard_41164.json"
        },
    ],
]


# The following list is to demonstrate how 'View' button works. Data is treated
#   differently in the actual code, but the resulting format will be similar.
quant_calib_json = [
    yaml.dump(quant_calib[0][0], default_flow_style=False, sort_keys=False, indent=4),
    yaml.dump(quant_calib[1][0], default_flow_style=False, sort_keys=False, indent=4),
]



# Use this file if you need to import PyXRF APIs into a custom script.
# Use 'pyxrf.api' if you are interactively importing APIs into an IPython session.

from .model.fileio import stitch_fitted_results  # noqa: F401
from .model.fileio import spec_to_hdf  # noqa: F401
from .model.fileio import create_movie  # noqa: F401
from .model.fileio import combine_data_to_recon  # noqa: F401
from .model.fileio import h5file_for_recon  # noqa: F401
from .model.fileio import export_to_view  # noqa: F401
from .model.fileio import make_hdf_stitched  # noqa: F401

from .model.load_data_from_db import make_hdf, export1d  # noqa: F401
from .model.command_tools import fit_pixel_data_and_save, pyxrf_batch  # noqa: F401
from .xanes_maps.xanes_maps_api import build_xanes_map  # noqa: F401
from .simulation.sim_xrf_scan_data import gen_hdf5_qa_dataset, gen_hdf5_qa_dataset_preset_1  # noqa: F401
from .core.map_processing import dask_client_create  # noqa: F401

from .model.load_data_from_db import save_data_to_hdf5  # noqa: F401, E402
from .model.fileio import read_data_from_hdf5  # noqa: F401, E402

# Note:  the statement '# noqa: F401' is telling flake8 to ignore violation F401 at the given line
#     Violation F401 - the package is imported but unused

import logging

logger = logging.getLogger("pyxrf")

logger.setLevel(logging.INFO)

formatter = logging.Formatter(fmt="%(asctime)s : %(levelname)s : %(message)s")

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.INFO)
logger.addHandler(stream_handler)

try:
    from .model.load_data_from_db import db
except ImportError:
    db = None
    logger.error("Databroker is not available.")

try:
    from .model.load_data_from_db import db_analysis
except ImportError:
    db_analysis = None
    # We don't use 'analysis' databroker, so disable the message for now
    # logger.error("'Analysis' databroker is not available.")



import os
import numpy as np

from qtpy.QtWidgets import (
    QPushButton,
    QHBoxLayout,
    QVBoxLayout,
    QGroupBox,
    QLabel,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QGridLayout,
)
from qtpy.QtGui import QIntValidator

from .useful_widgets import LineEditReadOnly, LineEditExtended, set_tooltip

import logging

logger = logging.getLogger(__name__)


class DialogLoadMask(QDialog):
    """
    Dialog box for selecting spatial ROI and mask.

    Typical use:

    default_dir = "/home/user/data"
    n_rows, n_columns = 15, 20  # Some values

    # Values that are changed by the dialog box
    roi = (2, 3, 11, 9)
    use_roi = True
    mask_f_path = ""
    use_mask = False

    dlg = DialogLoadMask()
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)
    dlg.set_roi(row_start=roi[0], column_start=roi[1], row_end=roi[2], column_end=roi[3])
    dlg.set_roi_active(use_roi)

    dlg.set_default_directory(default_dir)
    dlg.set_mask_file_path(mask_f_path)
    dlg.set_mask_file_active(use_mask)

    if dlg.exec() == QDialog.Accepted:
        # If success, then read the values back. Discard changes if rejected.
        roi = dlg.get_roi()
        use_roi = dlg.get_roi_active()
        mask_f_path = dlg.get_mask_file_path()
        use_mask = dlg.get_mask_file_active()
    """

    def __init__(self):

        super().__init__()

        self._validation_enabled = False

        self.resize(500, 300)
        self.setWindowTitle("Load Mask or Select ROI")

        # Initialize variables used with ROI selection group
        self._n_rows = 0
        self._n_columns = 0
        self._roi_active = False
        self._row_start = -1
        self._column_start = -1
        self._row_end = -1
        self._column_end = -1
        # ... with Mask group
        self._mask_active = False
        self._mask_file_path = ""
        self._default_directory = ""

        # Fields for entering spatial ROI coordinates
        self.validator_rows = QIntValidator()
        self.validator_rows.setBottom(1)
        self.validator_cols = QIntValidator()
        self.validator_cols.setBottom(1)

        self.le_roi_start_row = LineEditExtended()
        self.le_roi_start_row.setValidator(self.validator_rows)
        self.le_roi_start_row.editingFinished.connect(self.le_roi_start_row_editing_finished)
        self.le_roi_start_row.textChanged.connect(self.le_roi_start_row_text_changed)
        self.le_roi_start_col = LineEditExtended()
        self.le_roi_start_col.setValidator(self.validator_cols)
        self.le_roi_start_col.editingFinished.connect(self.le_roi_start_col_editing_finished)
        self.le_roi_start_col.textChanged.connect(self.le_roi_start_col_text_changed)
        self.le_roi_end_row = LineEditExtended()
        self.le_roi_end_row.setValidator(self.validator_rows)
        self.le_roi_end_row.editingFinished.connect(self.le_roi_end_row_editing_finished)
        self.le_roi_end_row.textChanged.connect(self.le_roi_end_row_text_changed)
        self.le_roi_end_col = LineEditExtended()
        self.le_roi_end_col.setValidator(self.validator_cols)
        self.le_roi_end_col.editingFinished.connect(self.le_roi_end_col_editing_finished)
        self.le_roi_end_col.textChanged.connect(self.le_roi_end_col_text_changed)
        self._text_map_size_base = "   * Map size: "
        self.label_map_size = QLabel(self._text_map_size_base + "not set")

        # Group box for spatial ROI selection
        self.gb_roi = QGroupBox("Select ROI (in pixels)")
        set_tooltip(
            self.gb_roi,
            "Select rectangular <b>spatial ROI</b>. If <b>mask</b> is "
            "loaded, then ROI is applied to the masked data.",
        )
        self.gb_roi.setCheckable(True)
        self.gb_roi.toggled.connect(self.gb_roi_toggled)
        self.gb_roi.setChecked(self._roi_active)
        vbox = QVBoxLayout()
        grid = QGridLayout()
        grid.addWidget(QLabel("Start position(*):"), 0, 0)
        grid.addWidget(QLabel("row"), 0, 1)
        grid.addWidget(self.le_roi_start_row, 0, 2)
        grid.addWidget(QLabel("column"), 0, 3)
        grid.addWidget(self.le_roi_start_col, 0, 4)
        grid.addWidget(QLabel("End position(*):"), 1, 0)
        grid.addWidget(QLabel("row"), 1, 1)
        grid.addWidget(self.le_roi_end_row, 1, 2)
        grid.addWidget(QLabel("column"), 1, 3)
        grid.addWidget(self.le_roi_end_col, 1, 4)
        vbox.addLayout(grid)
        vbox.addWidget(self.label_map_size)
        self.gb_roi.setLayout(vbox)

        # Widgets for loading mask
        self.pb_load_mask = QPushButton("Load Mask ...")
        self.pb_load_mask.clicked.connect(self.pb_load_mask_clicked)
        self._le_load_mask_default_text = "select 'mask' file"
        self.le_load_mask = LineEditReadOnly(self._le_load_mask_default_text)

        # Group box for setting mask
        self.gb_mask = QGroupBox("Set mask")
        set_tooltip(
            self.gb_mask,
            "Load <b>mask</b> from file. Active pixels in the mask are "
            "represented by positive integers. If <b>spatial ROI</b> is "
            "selected, then it is applied to the masked data.",
        )
        self.gb_mask.setCheckable(True)
        self.gb_mask.toggled.connect(self.gb_mask_toggled)
        self.gb_mask.setChecked(self._mask_active)
        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_load_mask)
        hbox.addWidget(self.le_load_mask)
        self.gb_mask.setLayout(hbox)

        # Yes/No button box
        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.button_ok = self.button_box.button(QDialogButtonBox.Ok)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        vbox = QVBoxLayout()
        vbox.addWidget(self.gb_roi)
        vbox.addStretch(1)
        vbox.addWidget(self.gb_mask)
        vbox.addStretch(2)
        vbox.addWidget(self.button_box)
        self.setLayout(vbox)

        self._validation_enabled = True
        self._validate_all_widgets()

    def _compute_home_directory(self):
        dir_name = "."
        if self._default_directory:
            dir_name = self._default_directory
        if self._mask_file_path:
            d, _ = os.path.split(self._mask_file_path)
            dir_name = d if d else dir_name
        return dir_name

    def pb_load_mask_clicked(self):
        dir_name = self._compute_home_directory()
        file_name = QFileDialog.getOpenFileName(self, "Load Mask From File", dir_name, "All (*)")
        file_name = file_name[0]
        if file_name:
            self._mask_file_path = file_name
            self._show_mask_file_path()
            logger.debug(f"Mask file is selected: '{file_name}'")

    def gb_roi_toggled(self, state):
        self._roi_activate(state)

    def _read_le_roi_value(self, line_edit, v_default):
        """
        Attempt to read value from line edit box as int, return `v_default` if not successful.

        Parameters
        ----------
        line_edit: QLineEdit
            reference to QLineEdit object
        v_default: int
            default value returned if the value read from edit box is incorrect
        """
        try:
            val = int(line_edit.text())
            if val < 1:
                raise Exception()
        except Exception:
            val = v_default
        return val

    def le_roi_start_row_editing_finished(self):
        self._row_start = self._read_le_roi_value(self.le_roi_start_row, self._row_start + 1) - 1

    def le_roi_end_row_editing_finished(self):
        self._row_end = self._read_le_roi_value(self.le_roi_end_row, self._row_end)

    def le_roi_start_col_editing_finished(self):
        self._column_start = self._read_le_roi_value(self.le_roi_start_col, self._column_start + 1) - 1

    def le_roi_end_col_editing_finished(self):
        self._column_end = self._read_le_roi_value(self.le_roi_end_col, self._column_end)

    def le_roi_start_row_text_changed(self):
        self._validate_all_widgets()

    def le_roi_end_row_text_changed(self):
        self._validate_all_widgets()

    def le_roi_start_col_text_changed(self):
        self._validate_all_widgets()

    def le_roi_end_col_text_changed(self):
        self._validate_all_widgets()

    def gb_mask_toggled(self, state):
        self._mask_file_activate(state)

    def _validate_all_widgets(self):
        """
        Validate the values and state of all widgets, update the 'valid' state
        of the widgets and enable/disable Ok button.
        """

        if not self._validation_enabled:
            return

        # Check if all fields have valid input values
        def _check_valid_input(line_edit, flag_valid):
            val = self._read_le_roi_value(line_edit, -1)
            state = val > 0
            line_edit.setValid(state)
            flag_valid = flag_valid if state else False
            return val, flag_valid

        # Set all line edits to 'valid' state
        self.le_roi_start_row.setValid(True)
        self.le_roi_end_row.setValid(True)
        self.le_roi_start_col.setValid(True)
        self.le_roi_end_col.setValid(True)
        self.le_load_mask.setValid(True)

        flag_valid = True

        if self._roi_active:
            # Perform the following checks only if ROI group is active
            rs, flag_valid = _check_valid_input(self.le_roi_start_row, flag_valid)
            re, flag_valid = _check_valid_input(self.le_roi_end_row, flag_valid)
            cs, flag_valid = _check_valid_input(self.le_roi_start_col, flag_valid)
            ce, flag_valid = _check_valid_input(self.le_roi_end_col, flag_valid)

            # Check if start
            if (rs > 0) and (re > 0) and (rs > re):
                self.le_roi_start_row.setValid(False)
                self.le_roi_end_row.setValid(False)
                flag_valid = False
            if (cs > 0) and (ce > 0) and (cs > ce):
                self.le_roi_start_col.setValid(False)
                self.le_roi_end_col.setValid(False)
                flag_valid = False

        if self._mask_active:
            if not self._mask_file_path:
                self.le_load_mask.setValid(False)
                flag_valid = False

        self.button_ok.setEnabled(flag_valid)

    def set_image_size(self, *, n_rows, n_columns):
        """
        Set the image size. Image size is used to for input validation. When image
        size is set, the selection is reset to cover the whole image.

        Parameters
        ----------
        n_rows: int
            The number of rows in the image: (1..)
        n_columns: int
            The number of columns in the image: (1..)
        """
        if n_rows < 1 or n_columns < 1:
            raise ValueError(
                "DialogLoadMask: image size have zero rows or zero columns: "
                f"n_rows={n_rows} n_columns={n_columns}. "
                "Report the error to the development team."
            )

        self._n_rows = n_rows
        self._n_columns = n_columns

        self._row_start = 0
        self._row_end = n_rows
        self._column_start = 0
        self._column_end = n_columns

        self._show_selection(True)
        self._validate_all_widgets()

        self.validator_rows.setTop(n_rows)
        self.validator_cols.setTop(n_columns)
        # Set label
        self.label_map_size.setText(f"{self._text_map_size_base}{self._n_rows} rows, {self._n_columns} columns.")

    def _show_selection(self, visible):
        """visible: True - show values, False - hide values"""

        def _show_number(l_edit, value):
            if value > 0:
                l_edit.setText(f"{value}")
            else:
                # This would typically mean incorrect initialization of the dialog box
                l_edit.setText("")

        _show_number(self.le_roi_start_row, self._row_start + 1)
        _show_number(self.le_roi_start_col, self._column_start + 1)
        _show_number(self.le_roi_end_row, self._row_end)
        _show_number(self.le_roi_end_col, self._column_end)

    def set_roi(self, *, row_start, column_start, row_end, column_end):
        """
        Set the values of fields that define selection of the image region. First set
        the image size (`set_image_size()`) and then set the selection.

        Parameters
        ----------
        row_start: int
            The row number of the first pixel in the selection (0..n_rows-1).
            Negative (-1) - resets value to 0.
        column_start: int
            The column number of the first pixel in the selection (0..n_columns-1).
            Negative (-1) - resets value to 0.
        row_end: int
            The row number following the last pixel in the selection (1..n_rows).
            This row is not included in the selection. Negative (-1) - resets value to n_rows.
        column_end: int
            The column number following the last pixel in the selection (1..n_columns).
            This column is not included in the selection. Negative (-1) - resets value to n_columns.
        """
        # The variables holding the selected region are following Python conventions for the
        #   selections: row_start, column_start are in the range 0..n_rows-1, 0..n_columns-1
        #   and row_end, column_end are in the range 1..n_rows, 1..n_columns.
        #   The values displayed in the dialog box are just pixels numbers in the range
        #   1..n_rows, 1..n_columns that define the rectangle in the way intuitive to the user.
        self._row_start = int(np.clip(row_start, a_min=0, a_max=self._n_rows - 1))
        self._column_start = int(np.clip(column_start, a_min=0, a_max=self._n_columns - 1))

        def _adjust_last_index(index, n_elements):
            if index < 0:
                index = n_elements
            index = int(np.clip(index, 1, n_elements))
            return index

        self._row_end = _adjust_last_index(row_end, self._n_rows)
        self._column_end = _adjust_last_index(column_end, self._n_columns)
        self._show_selection(self._roi_active)
        self._validate_all_widgets()

    def _roi_activate(self, state):
        self._roi_active = state
        self._show_selection(state)
        self._validate_all_widgets()

    def set_roi_active(self, state):
        self._roi_activate(state)
        self.gb_roi.setChecked(self._roi_active)

    def get_roi(self):
        return self._row_start, self._column_start, self._row_end, self._column_end

    def get_roi_active(self):
        return self._roi_active

    def _show_mask_file_path(self):
        fpath = self._mask_file_path if self._mask_file_path else self._le_load_mask_default_text
        self.le_load_mask.setText(fpath)
        self._validate_all_widgets()

    def _mask_file_activate(self, state):
        self._mask_active = state
        self._show_mask_file_path()
        self._validate_all_widgets()

    def set_mask_file_active(self, state):
        self._mask_file_activate(state)
        self.gb_mask.setChecked(self._mask_active)

    def get_mask_file_active(self):
        return self._mask_active

    def set_mask_file_path(self, fpath):
        self._mask_file_path = fpath
        self._show_mask_file_path()

    def get_mask_file_path(self):
        return self._mask_file_path

    def set_default_directory(self, dir_name):
        self._default_directory = dir_name

    def get_default_directory(self):
        return self._default_directory
